# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import re
from datetime import timedelta

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message
from kudubot.resources.images.__init__ import get_location as get_resource


class SimpleCommandsService(Service):
    """
    The SimpleCommandsService Class that extends the generic Service class.
    The service offers several trivial commands to the user
    """

    identifier = "simple_commands"
    """
    The identifier for this service
    """

    help_description = {"en": "Simple Commands: Collection of Commands that do simple things\n"
                              "/uptime\tDisplays the uptime of the server running the bot\n"
                              "/sl\tlike ls, but different.",
                        "de": "Simple Befehle: Sammlung von Befehlen die simple Sachen machen\n"
                              "/upzeit\tZeigt die 'uptime' des Servers auf dem der Bot läuft\n"
                              "/sl\twie ls, aber anders"}
    """
    Help description for this service.
    """

    commands = {"/uptime": "en",
                "/upzeit": "de",
                "/sl": "en"}
    """
    List of commands together with the language they are in and the method they call (initialized in initialize())
    """

    uptime_no_uptime_file = {"en": "Uptime command unavailable on this platform",
                             "de": "Upzeit Befehl auf dieser Platform nicht unterstützt"}

    def initialize(self) -> None:
        """
        Initializes the commands dictionary

        :return: None
        """
        self.commands = {"/uptime": ("en", self.get_uptime),
                         "/upzeit": ("de", self.get_uptime),
                         "/sl": ("en", self.steam_locomotive)}

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        self.connection.last_used_language = self.commands[message.message_body.lower()][0]
        command = self.commands[message.message_body.lower()][1]
        command(message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        regex = "^" + Service.regex_string_from_dictionary_keys([SimpleCommandsService.commands]) + "$"
        return re.search(re.compile(regex), message.message_body.lower())

    def get_uptime(self, message: Message) -> None:
        """
        Calculates the host PC's uptime and sends it to the sender

        :param message: The received message
        :return: None
        """
        if not os.path.isfile('/proc/uptime'):
            reply = self.uptime_no_uptime_file[self.connection.last_used_language]

        else:
            with open('/proc/uptime', 'r') as uptime_file:
                uptime_seconds = float(uptime_file.readline().split()[0])
                uptime_string = str(timedelta(seconds=uptime_seconds))
            reply = "Uptime: " + uptime_string

        reply_message = self.generate_reply_message(message, "Uptime", reply)
        self.send_text_message(reply_message)

    def steam_locomotive(self, message: Message) -> None:
        """
        Sends an image of a Steam locomotive to the sender

        :param message: the received message
        :return: None
        """
        image_file = get_resource("sl.jpg")
        self.send_image_message(message.address, image_file)



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message


class SimpleEqualsResponseService(Service):
    """
    The SimpleEqualsResponseService Class that extends the generic Service class.
    The service responds to strings that exactly match pre-defined options
    """

    identifier = "simple_equals_response"
    """
    The identifier for this service
    """

    help_description = {"en": "No Help Description Available",
                        "de": "Keine Hilfsbeschreibung verfügbar"}
    """
    Help description for this service. It's empty, because this service does not act on actual commands
    per say.
    """

    case_insensitive_options = {"uptime": "Much too long, I'm tired"}
    """
    Case-insensitive defined response conditions and responses
    """

    case_sensitive_options = {"ping": "pong",
                              "Ping": "Pong"}
    """
    Case-sensitive defined response conditions and responses
    """

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        try:
            reply = self.case_sensitive_options[message.message_body]
        except KeyError:
            reply = self.case_insensitive_options[message.message_body.lower()]

        reply_message = self.generate_reply_message(message, "Simple Equals Response", reply)
        self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        for option in SimpleEqualsResponseService.case_sensitive_options:
            if option == message.message_body:
                return True
        for option in SimpleEqualsResponseService.case_insensitive_options:
            if option == message.message_body.lower():
                return True
        return False



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import re
import random
import string

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message


class RandomKeyGeneratorService(Service):
    """
    The RandomKeyGeneratorService Class that extends the generic Service class.
    The service sends a random key of specified length
    """

    identifier = "random_key_generator"
    """
    The identifier for this service
    """

    help_description = {"en": "/randomkey\tgenerates a random key\n"
                              "syntax:\n"
                              "/randomkey <length>",
                        "de": "/zufallschlüssel\tgeneriert einen zufälligen Schlüssel\n"
                              "syntax:\n"
                              "/zufallschlüssel <Länge>"}
    """
    Help description for this service.
    """

    alphabet = string.ascii_letters + string.digits + string.punctuation
    """
    The alphabet to be used to generate a random key
    """

    random_key_keywords = {"/randomkey": "en",
                           "/zufallschlüssel": "de"}
    """
    Keywords for the /randomkey command
    """

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        language, length = message.message_body.lower().split(" ", 1)
        self.connection.last_used_language = self.random_key_keywords[language]

        reply = self.generate_key(int(length))
        reply_message = self.generate_reply_message(message, "Random Key", reply)
        self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        regex = "^" + Service.regex_string_from_dictionary_keys([RandomKeyGeneratorService.random_key_keywords]) \
                + " [1-9]{1}[0-9]*$"
        return re.search(re.compile(regex), message.message_body.lower())

    def generate_key(self, length: int) -> str:
        """
        Generates a random key of specified length using the alphabet specified as class variable

        :param length: the length of the keyphrase
        :return: the random key
        """
        random_key = ""
        if length > 100:
            return "Sorry"
        for x in range(0, length):
            random_key += random.choice(self.alphabet)
        return random_key



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import re
import time
import random
import datetime
from typing import List, Tuple, Set

from kudubot.connection.generic.Message import Message
from kudubot.config.LocalConfigChecker import LocalConfigChecker
from kudubot.services.local_services.CasinoService import CasinoService
from kudubot.resources.images.__init__ import get_location as get_board_image_file


def calculate_neighbour_group(roulette_board: List[List[int]]) -> List[Set[int]]:
    """
    Calculates the neighbouring tiles for a roulette board

    :param roulette_board: the roulette board to be used
    :return: the dual neghbours, the quad neighbours
    """
    board = roulette_board
    pairs = []
    quads = []
    row = 0
    while row < len(board):
        row_crawler = 0
        column_crawler = 1

        while column_crawler < len(board[row]):
            pairs.append([board[row][row_crawler], board[row][column_crawler]])

            if row < (len(board) - 1):
                pairs.append({board[row][row_crawler], board[row + 1][row_crawler]})
                pairs.append({board[row][column_crawler], board[row + 1][column_crawler]})

                quads.append({board[row][row_crawler],
                              board[row + 1][row_crawler],
                              board[row][column_crawler],
                              board[row + 1][column_crawler]})

                row_crawler += 1
            column_crawler += 1
        row += 1

    return pairs, quads


class RouletteService(CasinoService):
    """
    The RouletteService Class that extends the generic Service class.
    The service allows the user to play roulette.
    The board is spun every 2 minutes
    """

    identifier = "roulette"
    """
    The identifier for this service
    """

    has_background_process = True
    """
    Class has a background process.
    """

    help_description = {"en": "/roulette\tAllows the sender to play roulette\n"
                              "syntax: /roulette <amount> <bet>\n"
                              "valid bets: 0-36 | red | black | odd | even |"
                              " half 1-2 | group 1-3 | row 1-3 | batch 0-36{2|4}\n"
                              "/roulette time\tDisplays the time left until the wheel is spun again\n"
                              "/roulette board\tSends an image of a Roulette board as reference\n"
                              "/roulette bets\tDisplays all bets of a user\n"
                              "/roulette cancel <index>\tCancels a bet of a user\n"
                              "/roulette spin\tImmediately spins the wheel(admin)",
                        "de": "/roulette\tErmöglicht das Spielen von Roulette\n"
                              "syntax: /roulette <einsatz> <wette>\n"
                              "Mögliche Wetten:"
                              " 0-36 | rot | schwarz | ungerade | gerade | hälfte 1-2 | gruppe 1-3 | reihe 1-3 |"
                              "nachbarn 0-36{2|4}\n"
                              "/roulette zeit\t"
                              "Zeigt die verbleibende Zeit bis zum nächsten Drehen des Roulette-Rads an.\n"
                              "/roulette bord\tSchickt ein Bild eines Roulettebretts als Referenz\n"
                              "/roulette wetten\tZeigt alle Wetten eines Nutzers an\n"
                              "/roulette abbruch <index>\tBricht eine Wette des Nutzers ab\n"
                              "/roulette drehen\tDreht das Rouletterad(admin)"}
    """
    Help description for this service.
    """

    red_numbers = [1, 3, 5, 7, 9, 12, 14, 16, 18, 19, 21, 23, 25, 27, 30, 32, 34, 36]
    """
    List of numbers on the board that are red
    """

    black_numbers = [2, 4, 6, 8, 10, 11, 13, 15, 17, 20, 22, 24, 26, 28, 29, 31, 33, 35]
    """
    List of numbers on the board that are black
    """

    roulette_board = [[3, 6, 9, 12, 15, 18, 21, 24, 27, 30, 33, 36],
                      [2, 5, 8, 11, 14, 17, 20, 23, 26, 29, 32, 35],
                      [1, 4, 7, 10, 13, 16, 19, 22, 25, 28, 31, 34]]
    """
    The layout of the roulette board
    """

    dual_neighbours, quad_neighbours = calculate_neighbour_group(roulette_board)
    """
    lists of neighbouring tiles
    """

    roulette_directory = ""
    """
    Directory to store the roulette bets
    """

    roulette_board_image = get_board_image_file("rouletteboard.jpg")
    """
    The roulette board image file
    """

    spin_keywords = {"spin": "en",
                     "drehen": "de"}
    """
    Keywords for the 'spin' option
    """

    cancel_keywords = {"cancel": "en",
                       "abbruch": "de"}
    """
    Keywords for the 'cancel' option
    """

    board_keywords = {"board": "en",
                      "bord": "de"}
    """
    Keywords for the 'board' option
    """

    time_keywords = {"time": "en",
                     "zeit": "de"}
    """
    Keywords for the 'time' option
    """

    bets_keywords = {"bets": "en",
                     "wetten": "de"}
    """
    Keywords for the 'bets' option
    """

    black_keywords = {"black": "en",
                      "schwarz": "de"}
    """
    Keywords for the 'black' option
    """

    red_keywords = {"red": "en",
                    "rot": "de"}
    """
    Keywords for the 'red' option
    """

    even_keywords = {"even": "en",
                     "gerade": "de"}
    """
    Keywords for the 'even' option
    """

    odd_keywords = {"odd": "en",
                    "ungerade": "de"}
    """
    Keywords for the 'odd' option
    """

    group_keywords = {"group": "en",
                      "gruppe": "de"}
    """
    Keywords for the 'group' option
    """

    neighbour_keywords = {"neighbours": "en",
                          "nachbarn": "de"}
    """
    Keywords for the 'neighbour' option
    """

    row_keywords = {"row": "en",
                    "reihe": "de"}
    """
    Keywords for the 'row' option
    """

    half_keywords = {"half": "en",
                     "hälfte": "de"}
    """
    Keywords for the 'half' option
    """

    is_spinning_message = {"en": "The wheel is currently spinning, please wait a few seconds",
                           "de": "Das Rouletterad dreht sich gerade, hab einen moment Geduld"}
    """
    Message sent to the user when the roulette wheel is currently
    """

    won_message = {"en": " won ",
                   "de": " hat gewonnen: "}
    """
    Message shown when the user has won.
    """

    time_left_message = {"en": "s left until the board will be spinned",
                         "de": "s noch bis zum Drehen des Rades"}
    """
    Message to be sent when a request for time to next spin was sent
    """

    not_authenticated_method = {"en": "You are not authorized to spin the wheel",
                                "de": "Du hast keine Erlaubniss das Rad zu drehen"}
    """
    Message sent when a non-admin user tries to spin the wheel
    """

    undefined_behaviour = {"en": "This should not happen",
                           "de": "Dies sollte nicht passieren"}
    """
    Message for undefined behavour
    """

    spin_summary = {"en": "The result of the spin is: ",
                    "de": "Das Resultat ist: "}

    def initialize(self) -> None:
        """
        Initializes the roulette bets directory

        :return: None
        """
        super().initialize()
        self.roulette_directory = os.path.join(self.bet_directory, "roulette")
        LocalConfigChecker.validate_directory(self.roulette_directory)

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        reply = self.parse_user_input(message)

        if reply:
            reply_message = self.generate_reply_message(message, "Roulette", reply)
            self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        neighbours = RouletteService.regex_string_from_dictionary_keys([RouletteService.neighbour_keywords])
        group = RouletteService.regex_string_from_dictionary_keys([RouletteService.group_keywords])
        half = RouletteService.regex_string_from_dictionary_keys([RouletteService.half_keywords])
        row = RouletteService.regex_string_from_dictionary_keys([RouletteService.row_keywords])
        zero_to_36_regex = "(([0-9]?[0-3])|3[0-6])"

        regex = "^/roulette (([0-9]+((\.|,)[0-9]{2})?) (" + zero_to_36_regex + "|"
        regex += RouletteService.regex_string_from_dictionary_keys([RouletteService.red_keywords,
                                                                    RouletteService.black_keywords,
                                                                    RouletteService.odd_keywords,
                                                                    RouletteService.even_keywords])
        regex += "|" + neighbours + " " + zero_to_36_regex + "(-" + zero_to_36_regex + "){3}|"
        regex += "|" + neighbours + " " + zero_to_36_regex + "-" + zero_to_36_regex + "|"
        regex += "|" + group + " (1|2|3)|"
        regex += "|" + half + "(1|2)|"
        regex += "|" + row + "(1|2|3))|" \
                 + RouletteService.regex_string_from_dictionary_keys([RouletteService.spin_keywords,
                                                                      RouletteService.board_keywords,
                                                                      RouletteService.time_keywords,
                                                                      RouletteService.bets_keywords]) + "|"
        regex += RouletteService.regex_string_from_dictionary_keys([RouletteService.cancel_keywords]) + " [0-9]+)$"

        match = re.search(re.compile(regex), message.message_body.lower())

        if match and message.message_body.lower().split(" ")[1] in RouletteService.neighbour_keywords:

            neighbours = message.message_body.lower().split(" ")[2]
            neighbour_set = RouletteService.parse_neighbour_group(neighbours)

            match = neighbour_set in RouletteService.dual_neighbours \
                or neighbour_set in RouletteService.quad_neighbours

        return match

    def parse_user_input(self, message: Message) -> str:
        """
        Parses the user input from the message

        :param message: The message to be parsed
        :return: a reply string
        """
        user_input = message.message_body.lower()
        user_address = message.get_individual_address()

        user_input = user_input.split("/roulette ")[1]
        options = user_input.split(" ")

        if self.is_spinning():
            return self.is_spinning_message[self.connection.last_used_language]

        if options[0] in self.spin_keywords:
            if self.connection.authenticator.is_from_admin(message):
                return self.spin_wheel(self.spin_keywords[options[0]])
            else:
                return self.not_authenticated_method[self.spin_keywords[options[0]]]
        elif options[0] in self.cancel_keywords:
            return self.cancel_bets(self.cancel_keywords[options[0]], user_address, int(options[1]))
        elif options[0] in self.board_keywords:
            return self.send_board(user_address)
        elif options[0] in self.time_keywords:
            return str(self.request_time_to_spin()) + self.time_left_message[self.time_keywords[options[0]]]
        elif options[0] in self.bets_keywords:
            return self.get_bets_as_formatted_string("roulette", user_address)
        else:
            money = self.parse_money_string(options[0])

            if options[1] in self.red_keywords:
                selection = self.red_numbers
            elif options[1] in self.black_keywords:
                selection = self.black_numbers
            elif options[1] in self.odd_keywords:
                selection = []
                for number in (self.black_numbers + self.red_numbers):
                    if number % 2 == 1:
                        selection.append(number)
            elif options[1] in self.even_keywords:
                selection = []
                for number in (self.black_numbers + self.red_numbers):
                    if number % 2 == 0:
                        selection.append(number)
            elif re.search(r"(([0-9]?[0-3])|3[0-6])", options[1]):
                selection = [int(options[1])]
            else:
                parameter = options[2]

                if options[1] in self.neighbour_keywords:
                    selection = self.parse_neighbour_group(parameter)
                elif options[1] in self.group_keywords:
                    selection = []
                    start_point = (int(options[1]) - 1) * 12
                    for tile in range(start_point, start_point + 12):
                        selection.append(tile + 1)
                elif options[1] in self.half_keywords:
                    selection = []
                    start_point = (int(options[1]) - 1) * 18
                    for tile in range(start_point, start_point + 18):
                        selection.append(tile + 1)
                elif options[1] in self.row_keywords:
                    selection = self.roulette_board[int(options[1])]
                else:
                    return self.undefined_behaviour[self.connection.last_used_language]

            return self.create_bet(user_address, selection, money)

    @staticmethod
    def parse_neighbour_group(neighbour_group_string: str) -> Set[int]:
        """
        Parses a neighbour-group string and returns it as a set of tiles

        :param neighbour_group_string: the string to parse
        :return: a set of tiles
        """
        neighbours = neighbour_group_string.split("-")
        neighbour_set = set()

        for neigbour in neighbours:
            neighbour_set.add(int(neigbour))

        return neighbour_set

    @staticmethod
    def get_position_on_table(tile: int) -> Tuple[int, int]:
        """
        Gets the coordinates of a tile on the roulette table

        :param tile: the tile to check
        :return: the column, the row
        """
        return int(tile / 3), (2 - (tile - 1) % 3)

    @staticmethod
    def is_spinning() -> bool:
        """
        Checks if the board is spinning/should be spinning (which is a 5 second
        timeframe to avoid racing conditions)

        :return: True if the wheel should be spinning, False otherwise
        """
        current_time = datetime.datetime.utcnow()
        minutes = int(current_time.minute)
        seconds = int(current_time.second)

        return minutes % 2 == 1 and seconds >= 55

    def spin_wheel(self, language: str) -> None:
        """
        Spins the roulette wheel, calculating winnings for all users

        :param language: the languag to use
        :return: None
        """
        result = random.randint(0, 36)
        users = []

        for bet_file_name in os.listdir(self.roulette_directory):
            user = bet_file_name.rsplit("###BETVAL=", 1)[0]
            if user not in users:
                users.append(user)

        bets = []
        addresses = []

        for user in users:
            user_bets = self.get_bets("roulette", user)

            for bet in user_bets:
                if bet["address"] not in addresses:
                    addresses.append(bet["address"])

                selected_tiles, multiplier = bet['bet_type'].split("X")
                selected_tiles = selected_tiles.split("-")

                bet['multiplier'] = int(multiplier)
                bet['selection'] = []

                for tile in selected_tiles:
                    bet['selection'].append(int(tile))

                os.remove(bet["file"])
            bets += user_bets

        summary = ""

        for bet in bets:
            if bet["address"] not in addresses:
                addresses.append(bet["address"])

            if result in bet["selection"]:
                won_value = bet["value"] * bet["multiplier"]
                summary += bet["user"] + self.won_message[language] + self.format_money_string(won_value) + "\n\n"
                self.transfer_funds(bet["user"], won_value)

        if len(addresses) > 0:
            summary = self.spin_summary[language] + str(result) + "\n\n" + summary

            for better in addresses:
                message = Message(message_body=summary, message_title="Roulette", address=better)
                self.send_text_message(message)

    def cancel_bets(self, language: str, user_address: str, index: int) -> str:
        """
        Cancels a bet

        :param language: the language to use
        :param user_address: the user address string
        :param index: the index of the bet to delete
        :return: A message detailing the success or failure of the deletion
        """
        self.connection.last_used_language = language
        return self.delete_bet("roulette", user_address, index)

    def send_board(self, address: str) -> None:
        """
        Sends an image of a roulette board to the user

        :param address: the address of the user
        :return: None
        """
        self.send_image_message(address, self.roulette_board_image)

    @staticmethod
    def request_time_to_spin() -> int:
        """
        Returns the mount of time in seconds until the roulette wheel is spun again

        :return: the time until the next spin in seconds
        """
        current_time = datetime.datetime.utcnow()
        current_minute = int(current_time.minute)
        current_second = int(current_time.second)

        time_left = 120 - current_second - 5
        if current_minute % 2 == 1:
            time_left -= 60

        return time_left

    def create_bet(self, address: str, bet_items: List[int], value: int) -> str:
        """
        Creates a new bet

        :param address: The address used to determine which user made the bet
        :param bet_items: The items for which the user placed a bet
        :param value: the value of the bet
        :return: a message telling the user that the bet was placed succesfully
        """
        multiplier = int(36 / len(bet_items))
        bet_string = ""

        for element in bet_items:
            bet_string += str(element) + "-"

        bet_string = bet_string.rsplit("-", 1)[0]
        bet_string += "X" + str(multiplier)
        reply = self.store_bet("roulette", address, value, bet_string)

        return reply

    def background_process(self) -> None:
        """
        Spins the wheel in periodic intervals

        :return: None
        """
        while True:
            if self.is_spinning():
                self.spin_wheel(self.connection.last_used_language)
                time.sleep(60)
            time.sleep(3)



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os

from kudubot.config.LocalConfigChecker import LocalConfigChecker
from kudubot.connection.generic.Message import Message


class Authenticator(object):
    """
    Class that handles the permission leves for various users
    """

    admin_contact_file = ""
    """
    The contact file to use to determine the admin users
    """

    blacklisted_contact_file = ""
    """
    The contact file to use to determine the blacklisted users
    """

    def __init__(self, connection_type: str) -> None:
        """
        Creates a new Authenticator for a specific connection type

        :param connection_type: The connection type used
        :return: None
        """
        self.admin_contact_file = os.path.join(LocalConfigChecker.contacts_directory, connection_type, "admin")
        self.blacklisted_contact_file = \
            os.path.join(LocalConfigChecker.contacts_directory, connection_type, "blacklist")

    @staticmethod
    def is_in_file(user: str, contact_file: str) -> bool:
        """
        Checks if a user is in a conatct file or not

        :param user: The user to search for
        :param contact_file: the contact file to search
        :return: True if the user is in the contact file, False otherwise
        """
        contacts = open(contact_file, 'r')
        content = contacts.read()
        contacts.close()

        return user in content.split("\n")

    def is_from_admin(self, message: Message) -> bool:
        """
        Checks if a message originated from an admin user

        :param message: the message to be checked
        :return: True if admin, False if not
        """
        identifier = message.get_individual_address()

        return Authenticator.is_in_file(identifier, self.admin_contact_file)

    def is_from_blacklisted_user(self, message: Message) -> bool:
        """
        Checks if a message originated from a blacklisted user

        :param message: the message to be checked
        :return: True if blacklisted, False if not
        """
        identifier = message.get_individual_address()

        return Authenticator.is_in_file(identifier, self.blacklisted_contact_file)



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import kudubot.metadata as metadata


class PrintLogger(object):
    """
    Class that handles printing to the console
    """

    @staticmethod
    def print(string: str, verbosity_level: int = 0) -> None:
        """
        Prints a string to the console while catching OSErrors, which can occur if no terminal is available,
        for example after exiting an SSH session

        :param string: the string to print
        :param verbosity_level: after which verbosity level the string is printed
        :return: None
        """
        if metadata.verbosity >= verbosity_level:
            try:
                print(string)
            except OSError:
                pass



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import re

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message


class HelloWorldService(Service):
    """
    The HelloWorldService Class that extends the generic Service class.
    It sends a Hello World code snipped in the requested programming language
    """

    identifier = "hello_world"
    """
    The identifier for this service
    """

    help_description = {"en": "/helloworld\tSends a message containing how to write a 'Hello World'"
                              "Program in a specific language\n"
                              "syntax:\n"
                              "/helloworld <language|list>",
                        "de": "/helloworld\tSchickt eine Nachricht mit einem 'Hello World' Codeschnipsel für eine"
                              "spezifische Programmiersprache\n"
                              "syntax:\n"
                              "/helloworld <sprache|liste>"}
    """
    Help description for this service.
    """

    implementations = {"python": "print(\"Hello World!\")",
                       "python2": "print \"Hello World!\"",
                       "python3": "print(\"Hello World!\")",
                       "java": "pubic class Main {\n"
                               "    public static void main(String[] args) {\n"
                               "        System.out.println(\"Hello World!\");\n"
                               "    }\n"
                               "}",
                       "c": "#include <stdio.h>\n\n"
                            "int main() {\n"
                            "    printf(\"Hello World!\\n\");\n"
                            "    return 0;\n"
                            "}",
                       "c++": "#include <iostream>\n\n"
                              "int main() {\n"
                              "    std::cout << \"Hello World!\";\n"
                              "    return 0;\n"
                              "}",
                       "bash": "echo \"Hello World\"",
                       "rust": "fn main() {\n"
                               "    println!(\"Hello World!\");\n"
                               "}",
                       "ruby": "puts 'Hello, world!'",
                       "perl": "print \"Hello World!\\n\";",
                       "c#": "using System;\n"
                             "namespace HelloWorldApplication{\n"
                             "    class HelloWorld {\n"
                             "        Console.WriteLine(\"Hello World!\");\n"
                             "        Console.ReadKey();\n"
                             "    }\n"
                             "}",
                       "basic": "10 PRINT \"Hello World!\"",
                       "visual basic": "Module HelloWorld\n"
                                       "    Sub Main()\n"
                                       "        System.Console.WriteLine(\"Hello World!\")\n"
                                       "        System.Console.ReadLine()\n"
                                       "        End\n"
                                       "    End Sub\n"
                                       "End Module\n",
                       "brainfuck": "++++++++++\n"
                                    "[\n"
                                    " >+++++++>++++++++++>+++>+<<<<-\n"
                                    "]\n"
                                    ">++.\n"
                                    ">+.\n"
                                    "+++++++.\n"
                                    ".\n"
                                    "+++.\n"
                                    ">++.\n"
                                    "<<+++++++++++++++.\n"
                                    ">.\n"
                                    "+++.\n"
                                    "------.\n"
                                    "--------.\n"
                                    ">+.\n"
                                    ">.\n"
                                    "+++.",
                       "haskell": "main = putStrLn \"Hello World!\"",
                       "erlang": "-module(hello).\n"
                                 "-export([hello_world/0]).\n\n"
                                 "hello_world() -> io:fwrite(\"Hello World!\n\").",
                       "prolog": "?- write('Hello World!'), nl.",
                       "swift": "print(\"Hello World!\")",
                       "b": "main() {\n"
                            "    printf(\"Hello World!\");\n"
                            "}",
                       "d": "import std.stdio;\n"
                            "void main() {\n"
                            "    writeln(\"Hello World!\");\n"
                            "}",
                       "cobol": "000100 IDENTIFICATION DIVISION.\n"
                                "000200 PROGRAM-ID. HELLOWORLD.\n"
                                "000900 PROCEDURE DIVISION.\n"
                                "001000 MAIN.\n"
                                "001100 DISPLAY \"Hello World!\".\n"
                                "001200 STOP RUN.",
                       "fortran": "program hello\n"
                                  "write(*,*) \"Hello World!\"\n"
                                  "end program hello",
                       "go": "package main\n\n"
                             "import \"fmt\"\n\n"
                             "func main() {\n"
                             "    fmt.Println(\"Hello World!\")\n"
                             "}",
                       "lua": "print (\"Hello World!\")",
                       "x86 assembly": "section .data\n"
                                       "str:     db 'Hello World!', 0Ah\n"
                                       "str_len: equ $ - str\n\n\n"
                                       "section .text\n"
                                       "global _start\n\n"
                                       "_start:\n"
                                       "    mov eax, 4\n"
                                       "    mov ebx, 1\n\n"
                                       "    mov ecx, str\n"
                                       "    mov edx, str_len\n"
                                       "    int 80h\n\n"
                                       "    mov eax, 1\n"
                                       "    mov ebx, 0\n"
                                       "    int 80h",
                       "shakespeare": "Romeo, a young man with a remarkable patience.\n"
                                       "Juliet, a likewise young woman of remarkable grace.\n"
                                       "Ophelia, a remarkable woman much in dispute with Hamlet.\n"
                                       "Hamlet, the flatterer of Andersen Insulting A/S.\n",
                       "0815": "<:48:x<:65:=<:6C:$=$=$$~<:03:+\n"
                               "$~<:ffffffffffffffb1:+$<:77:~$\n"
                               "~<:fffffffffffff8:x+$~<:03:+$~\n"
                               "<:06:x-$x<:0e:x-$=x<:43:x-$",
                       "blub": "blub. blub? blub. blub. blub. blub. blub. blub. blub. blub. blub. blub. blub. blub. "
                               "blub. blub. blub. blub. blub. blub. blub! blub?\n"
                               "blub? blub. blub. blub. blub. blub. blub. blub. blub. blub. blub. blub. blub. blub. "
                               "blub. blub. blub. blub. blub. blub? blub! blub!\n"
                               "blub? blub! blub? blub. blub! blub. blub. blub? blub. blub. blub. blub. blub. blub. "
                               "blub. blub. blub. blub. blub. blub. blub. blub.\n"
                               "blub! blub? blub? blub. blub. blub. blub. blub. blub. blub. blub. blub. blub. blub? "
                               "blub! blub! blub? blub! blub? blub. blub. blub.\n"
                               "blub! blub. blub. blub. blub. blub. blub. blub. blub. blub. blub. blub. blub. blub. "
                               "blub. blub. blub! blub. blub! blub. blub. blub.\n"
                               "blub. blub. blub. blub. blub! blub. blub. blub? blub. blub? blub. blub? blub. blub. "
                               "blub. blub. blub. blub. blub. blub. blub. blub.\n"
                               "blub. blub. blub. blub. blub. blub. blub! blub? blub? blub. blub. blub. blub. blub. "
                               "blub. blub. blub. blub. blub. blub? blub! blub!\n"
                               "blub? blub! blub? blub. blub! blub. blub. blub? blub. blub? blub. blub? blub. blub. "
                               "blub. blub. blub. blub. blub. blub. blub. blub.\n"
                               "blub. blub. blub. blub. blub. blub. blub. blub. blub. blub. blub! blub? blub? blub. "
                               "blub. blub. blub. blub. blub. blub. blub. blub.\n"
                               "blub. blub. blub. blub. blub. blub. blub. blub. blub. blub. blub. blub? blub! blub! "
                               "blub? blub! blub? blub. blub! blub! blub! blub!\n"
                               "blub! blub! blub! blub. blub? blub. blub? blub. blub? blub. blub? blub. blub! blub. "
                               "blub. blub. blub. blub. blub. blub. blub! blub.\n"
                               "blub! blub! blub! blub! blub! blub! blub! blub! blub! blub! blub! blub! blub! blub. "
                               "blub! blub! blub! blub! blub! blub! blub! blub!\n"
                               "blub! blub! blub! blub! blub! blub! blub! blub! blub! blub. blub. blub? blub. blub?"
                               "blub. blub. blub! blub."}
    """
    The actual implementations of hello world in the different languages
    """

    language_not_found_error = {"en": "Programming Language not found",
                                "de": "Programmiersprache nicht gefunden"}
    """
    Error message sent when the programming language could not be found.
    """

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        prog_language = message.message_body.lower().split(" ", 1)[1]
        if prog_language.startswith("list"):
            reply = self.list_languages()
        else:
            reply = self.implementations[prog_language]

        reply_message = self.generate_reply_message(message, "Hello World", reply)

        if self.connection.identifier in ["whatsapp", "telegram"] and not prog_language.startswith("list"):
            self.send_text_as_image_message(reply_message)
        else:
            self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        regex = "^/helloworld (list(e)?|" \
                + Service.regex_string_from_dictionary_keys([HelloWorldService.implementations]) + ")$"
        regex = regex.replace("+", "\+")
        return re.search(re.compile(regex), message.message_body.lower())

    def list_languages(self) -> str:
        """
        Creates a list of implemented languages

        :return: the list of implemented languages
        """
        list_string = ""
        for language in self.implementations:
            list_string += language + "\n"
        return list_string.rstrip("\n")



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import re
from gtts import gTTS
from typing import Tuple

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message
from kudubot.config.LocalConfigChecker import LocalConfigChecker


class GoogleTtsService(Service):
    """
    The GoogleTtsService Class that extends the generic Service class.
    The service allows the user to send voice messages generated using Google's text to speech engine
    """

    identifier = "google_text_to_speech"
    """
    The identifier for this service
    """

    help_description = {"en": "/say\tA text-to-speech engine\n"
                              "syntax:\n"
                              "/say \"<text>\" [<language>]",
                        "de": "/sag\tEine text-to-speech Funktion\n"
                              "syntax:\n"
                              "/sag \"<text>\" [<sprache>]"}
    """
    Help description for this service.
    """

    supported_languages = {'af': 'Afrikaans',
                           'sq': 'Albanian',
                           'ar': 'Arabic',
                           'hy': 'Armenian',
                           'ca': 'Catalan',
                           'zh': 'Chinese',
                           'zh-cn': 'Chinese (Mandarin/China)',
                           'zh-tw': 'Chinese (Mandarin/Taiwan)',
                           'zh-yue': 'Chinese (Cantonese)',
                           'hr': 'Croatian',
                           'cs': 'Czech',
                           'da': 'Danish',
                           'nl': 'Dutch',
                           'en': 'English',
                           'en-au': 'English (Australia)',
                           'en-uk': 'English (United Kingdom)',
                           'en-us': 'English (United States)',
                           'eo': 'Esperanto',
                           'fi': 'Finnish',
                           'fr': 'French',
                           'de': 'German',
                           'el': 'Greek',
                           'ht': 'Haitian Creole',
                           'hi': 'Hindi',
                           'hu': 'Hungarian',
                           'is': 'Icelandic',
                           'id': 'Indonesian',
                           'it': 'Italian',
                           'ja': 'Japanese',
                           'ko': 'Korean',
                           'la': 'Latin',
                           'lv': 'Latvian',
                           'mk': 'Macedonian',
                           'no': 'Norwegian',
                           'pl': 'Polish',
                           'pt': 'Portuguese',
                           'pt-br': 'Portuguese (Brazil)',
                           'ro': 'Romanian',
                           'ru': 'Russian',
                           'sr': 'Serbian',
                           'sk': 'Slovak',
                           'es': 'Spanish',
                           'es-es': 'Spanish (Spain)',
                           'es-us': 'Spanish (United States)',
                           'sw': 'Swahili',
                           'sv': 'Swedish',
                           'ta': 'Tamil',
                           'th': 'Thai',
                           'tr': 'Turkish',
                           'vi': 'Vietnamese',
                           'cy': 'Welsh',
                           }
    """
    The languages supported by the text to speech engine
    """

    say_keywords = {"say": "en",
                    "sag": "de"}
    """
    Language keywords for the say command
    """

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        speech_text, language = self.parse_user_input(message.message_body.lower())
        audio_file = self.generate_audio(speech_text, language)
        self.send_audio_message(message.address, audio_file, caption=speech_text)
        self.delete_file_after(audio_file, 5)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        regex = "^/" + Service.regex_string_from_dictionary_keys([GoogleTtsService.say_keywords])
        regex += " \"[^\"]+\"( "
        regex += Service.regex_string_from_dictionary_keys([GoogleTtsService.supported_languages])
        regex += ")?$"

        return re.search(re.compile(regex), message.message_body.lower())

    @staticmethod
    def parse_user_input(user_input) -> Tuple[str, str]:
        """
        Parses the user input and determines the message text to send

        :param user_input: the input to be checked
        :return: a tuple of message text and the requested language
        """

        parts = user_input.split(" ", 1)
        say_key = parts.pop(0)
        language = GoogleTtsService.say_keywords[say_key.split("/")[1]]

        text_message = parts[0].split("\"", 2)[1]

        try:
            language = parts[0].rsplit("\" ", 1)[1]
        except IndexError:
            pass

        return text_message, language

    @staticmethod
    def generate_audio(text_string: str, language: str) -> str:
        """
        Generates an audio file using google's text to speech engine

        :param text_string: The string to be converted into speech
        :param language: The language to be used
        :return: the file path to the audio file
        """
        temp_file = os.path.join(LocalConfigChecker.program_directory, "tts_temp.mp3")
        Service.wait_until_delete(temp_file, 5)

        tts = gTTS(text=text_string, lang=language)
        tts.save(temp_file)
        return temp_file



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import re
import traceback

from yowsup.layers.interface import YowInterfaceLayer
from yowsup.layers.interface import ProtocolEntityCallback
from yowsup.layers.protocol_messages.protocolentities import TextMessageProtocolEntity

from kudubot.logger.PrintLogger import PrintLogger
from kudubot.connection.generic.Message import Message
from kudubot.logger.ExceptionLogger import ExceptionLogger
from kudubot.connection.generic.Connection import Connection
from kudubot.connection.whatsapp.layers.YowsupEchoLayer import YowsupEchoLayer
from kudubot.connection.whatsapp.stacks.YowsupEchoStack import YowsupEchoStack
from kudubot.connection.whatsapp.parsers.WhatsappConfigParser import WhatsappConfigParser


class WhatsappConnection(YowsupEchoLayer, Connection):
    """
    Class that implements the connection to the Whatsapp Messaging service
    """

    identifier = "whatsapp"
    """
    A string identifier with which other parts of the program can identify the type of connection
    """

    def __init__(self) -> None:
        """
        Constructor for the WhatsappConnection class that initializes the Yowsup layer

        :return: None
        """
        super().__init__()
        # noinspection PyCallByClass
        YowInterfaceLayer.__init__(self)
        self.initialize()

    def send_text_message(self, message: Message) -> None:
        """
        Sends a text message to the receiver. Some services allow the use of titles, but some don't,
        so the message title is optional

        :param message: The message to be sent
        :return: None
        """
        message_protocol_entity = self.convert_message_to_text_message_protocol_entity(message)
        self.toLower(message_protocol_entity)

    def send_image_message(self, receiver: str, message_image: str, caption: str = "") -> None:
        """
        Sends an image to the receiver, with an optional caption/title

        :param receiver: The receiver of the message
        :param message_image: The image to be sent
        :param caption: The caption/title to be displayed along with the image, defaults to an empty string
        :return: None
        """
        self.send_image(receiver, message_image, caption)

    def send_audio_message(self, receiver: str, message_audio: str, caption: str = "") -> None:
        """
        Sends an audio file to the receiver, with an optional caption/title

        :param receiver: The receiver of the message
        :param message_audio: The audio file to be sent
        :param caption: The caption/title to be displayed along with the audio, defaults to an empty string
        :return: None
        """
        str(caption)
        self.send_audio(receiver, message_audio)

    @staticmethod
    def establish_connection() -> None:
        """
        Establishes the connection to the specific service

        :return: None
        """
        credentials = WhatsappConfigParser.parse_whatsapp_config(WhatsappConnection.identifier)

        while True:
            try:
                PrintLogger.print("Starting Whatsapp Connection", 1)
                echo_stack = YowsupEchoStack(WhatsappConnection, credentials)
                echo_stack.start()
            except Exception as e:
                stack_trace = traceback.format_exc()
                ExceptionLogger.log_exception(e, stack_trace, "whatsapp")

    @ProtocolEntityCallback("message")
    def on_message(self, message_protocol_entity: TextMessageProtocolEntity):
        """
        Method run when a message is received
        :param message_protocol_entity: the message received
        :return: void
        """
        # Wrap the message protocol entity in a PEP8-compliant Wrapper
        if message_protocol_entity.getType() == "text":
            message = self.convert_text_message_protocol_entity_to_message(message_protocol_entity)
            self.on_incoming_message(message)

    @staticmethod
    def convert_text_message_protocol_entity_to_message(message_protocol_entity: TextMessageProtocolEntity) -> Message:
        """
        Converts an incoming text message protocol entity into a Message object

        :param message_protocol_entity: The entity to convert
        :return: the converted message
        """
        body = message_protocol_entity.getBody()

        sender_number = message_protocol_entity.getFrom(True)
        group_identifier = message_protocol_entity.getFrom(False)
        sender_name = message_protocol_entity.getNotify()
        group = False
        individual_number = ""
        individual_name = ""

        timestamp = float(message_protocol_entity.getTimestamp())

        if re.search(r"[0-9]+-[0-9]+", group_identifier):
            group = True
            individual_number = message_protocol_entity.getParticipant(True)
            individual_name = message_protocol_entity.getNotify()

        return Message(message_body=body, message_title="", address=sender_number, incoming=True, name=sender_name,
                       single_address=individual_number, single_name=individual_name, group=group, timestamp=timestamp)

    @staticmethod
    def convert_message_to_text_message_protocol_entity(message: Message) -> TextMessageProtocolEntity:
        """
        Converts an outgoing message object into a text message protocol entity

        :param message: The message to be converted
        :return: The converted text message protocol entity
        """
        to = message.address
        body = message.message_body

        return TextMessageProtocolEntity(body, to=to)



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import re
import requests
from bs4 import BeautifulSoup
from typing import Dict, List

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message


class FootballInfoService(Service):
    """
    The FootballInfoService Class that extends the generic Service class.
    The service parses www.livescore.com to get current league table and match results
    for football matches.
    """

    identifier = "football_info"
    """
    The identifier for this service
    """

    help_description = {"en": "/table\tSends football table information\n"
                              "syntax: /table [<country>][, <league>]\n\n"
                              "/matchday\tSends football matchday information\n"
                              "syntax: /matchday [<country>][, <league>]",
                        "de": "/tabelle\tSchickt Fußball Tabelleninformationen\n"
                              "syntax: /tabelle [<land>][, <liga>]\n\n"
                              "/spieltag\tSchickt Fußball Spieltaginformationen\n"
                              "syntax: /spieltag [<country>][, <liga>]"}
    """
    Help description for this service.
    """

    league_mode = False
    """
    If the command parser check the command, it can set this flag to tell the program that the league table is
    requested
    """

    matchday_mode = False
    """
    If the command parser check the command, it can set this flag to tell the program that the match day results
    are requested
    """

    league_descriptors = {"league": "en", "liga": "de"}
    """
    Descriptors in different languages for the league mode (used for the command syntax)
    """

    matchday_descriptors = {"matchday": "en", "spieltag": "de"}
    """
    Descriptors in different languages for the matchday mode (used for the command syntax)
    """

    country = ""
    """
    The country to parse
    """

    league = ""
    """
    The league to parse
    """

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        self.parse_command(message.message_body.lower())

        reply = ""

        if self.league_mode:
            reply = self.get_league_info()
        elif self.matchday_mode:
            reply = self.get_matchday_info()

        reply_message = self.generate_reply_message(message, "Football Info", reply)

        if self.connection.identifier in ["whatsapp", "telegram"]:
            self.send_text_as_image_message(reply_message)
        else:
            self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """

        # Generate the Regex
        regex_term = "^/" + Service.regex_string_from_dictionary_keys([FootballInfoService.league_descriptors,
                                                                       FootballInfoService.matchday_descriptors]) \
                     + "( [a-zA-Z]+, [a-zA-Z]+( |[a-zA-Z]+|-)*)[a-zA-Z]+$"

        return re.search(re.compile(regex_term), message.message_body.lower())

    def parse_command(self, message_text: str) -> None:
        """
        Parses the command. It determines which country and league to parse and if a league table or
        a match day. The language is also checked

        :return: None
        """
        # Check which mode to use
        mode = message_text.split(" ")[0].split("/")[1].lower()
        if mode in self.league_descriptors:
            self.connection.last_used_language = self.league_descriptors[mode]
            self.league_mode = True
        elif mode in self.matchday_descriptors:
            self.connection.last_used_language = self.matchday_descriptors[mode]
            self.matchday_mode = True

        # Determine league and country
        country_league = message_text.split(" ", 1)[1]
        self.country = country_league.split(", ")[0]
        self.league = country_league.split(", ")[1].replace(" ", "-")

    def get_league_info(self) -> str:
        """
        Retrieves the info for a league table, pretifies it, and returns it as a string

        :return: the league table info
        """
        teams = self.get_web_resource('.team')
        stats = self.get_web_resource('.pts')

        league_table = {}

        team_position_index = 1
        stats_index = 8

        while team_position_index < len(teams) and stats_index < len(stats):

            team = {"team_name": teams[team_position_index].text,
                    "matches": stats[stats_index].text,
                    "wins": stats[stats_index + 1].text,
                    "draws": stats[stats_index + 2].text,
                    "losses": stats[stats_index + 3].text,
                    "goals_for": stats[stats_index + 4].text,
                    "goals_against": stats[stats_index + 5].text,
                    "goal_difference": stats[stats_index + 6].text,
                    "points": stats[stats_index + 7].text,
                    "position": str(team_position_index)}

            league_table[team_position_index] = team

            team_position_index += 1
            stats_index += 8

        return self.format_league_table(league_table)

    def get_matchday_info(self) -> str:
        """
        Retrieves the info for a matchday, pretifies it, and returns it as a string

        :return: the matchday results info
        """
        teams = self.get_web_resource('.ply')
        times = self.get_web_resource('.min')
        score = self.get_web_resource('.sco')

        matches = []

        for index in range(0, len(teams), 2):
            match = {"left_team": teams[index].text,
                     "right_team": teams[index + 1].text,
                     "time": times[int(index / 2)].text,
                     "score": score[int(index / 2)].text}

            matches.append(match)

        return self.format_matchday(matches)

    def get_web_resource(self, html_element: str) -> List:
        """
        Gets the web resource for a specific HTML element from livescore.com
        This will be applied to whatever the currently selected country and league are.

        :param html_element: the html element to look for
        :return: the resource found
        """
        url = "http://www.livescore.com/soccer/" + self.country + "/" + self.league + "/"
        html = requests.get(url).text
        soup = BeautifulSoup(html, "html.parser")
        return soup.select(html_element)

    @staticmethod
    def format_league_table(league_table_dictionary: Dict[int, Dict[str, str]]) -> str:
        """
        Formats a league table dictionary

        :param league_table_dictionary: The league table dictionary to format
        :return: the league table as string
        """
        # Determine how long the longest team name is
        longest_team_name = 0
        for position in range(1, len(league_table_dictionary) + 1):
            team_name_length = len(league_table_dictionary[position]["team_name"])
            if team_name_length > longest_team_name:
                longest_team_name = team_name_length

        formatted_string = "#    " + "Team Name".ljust(longest_team_name) + \
                           "  P     W     D     L     GF    GA    GD    Pts\n"
        divider = "  "

        for position in range(1, len(league_table_dictionary) + 1):
            line = "\n" + str(position).ljust(3) + divider
            line += league_table_dictionary[position]["team_name"].ljust(longest_team_name) + divider
            line += league_table_dictionary[position]["matches"].ljust(4) + divider
            line += league_table_dictionary[position]["wins"].ljust(4) + divider
            line += league_table_dictionary[position]["draws"].ljust(4) + divider
            line += league_table_dictionary[position]["losses"].ljust(4) + divider
            line += league_table_dictionary[position]["goals_for"].ljust(4) + divider
            line += league_table_dictionary[position]["goals_against"].ljust(4) + divider
            line += league_table_dictionary[position]["goal_difference"].rjust(4) + divider
            line += league_table_dictionary[position]["points"].ljust(4)

            formatted_string += line

        return formatted_string

    @staticmethod
    def format_matchday(matchday_list: List[Dict[str, str]]) -> str:
        """
        Formats a matchday List into a sendable string

        :param matchday_list: The matchday list to be formatted
        :return: The formatted string
        """

        # Establish lengths of the team names
        longest_left = 0
        longest_right = 0

        for match in matchday_list:
            if len(match["left_team"]) > longest_left:
                longest_left = len(match["left_team"])
            if len(match["right_team"]) > longest_right:
                longest_right = len(match["left_team"])

        formatted_string = ""
        first = True

        for match in matchday_list:
            if first:
                first = False
            else:
                formatted_string += "\n"

            formatted_string += match["time"].ljust(6)
            formatted_string += match["left_team"].rjust(longest_left + 1)
            formatted_string += match["score"].ljust(5)
            formatted_string += match["right_team"].ljust(longest_right + 1)

        return formatted_string



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import re
import time
import datetime
from typing import Tuple, List, Dict

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message
from kudubot.config.LocalConfigChecker import LocalConfigChecker


class WeeklyReminderService(Service):
    """
    The WeeklyReminderService Class that extends the generic Service class.
    The service offers weekly reminder messages
    """

    identifier = "weekly_reminder"
    """
    The identifier for this service
    """

    help_description = {"en": "/w-remind\tStores a continuous (weekly) reminder\n"
                              "syntax:\n"
                              "/w-remind \"<message>\" <day> <hh-mm-ss>\n"
                              "/w-remind list\tLists all reminders currently stored"
                              "/w-remind delete <index>\tDeletes the reminder at the given index",
                        "de": "/w-remind\tSpeichert eine wöchentliche Errinnerung\n"
                              "syntax:\n"
                              "/w-erinner \"<nachricht>\" <tag> <hh-mm-ss>\n"
                              "/w-erinner liste\tListet alle Erinnerungen "
                              "/w-erinner löschen <index>\tLöscht die Erinnerung am gegebenen Index"}
    """
    Help description for this service.
    """

    has_background_process = True
    """
    Indicates that the serice has a background process
    """

    w_reminder_directory = "w_reminder"
    """
    The directory in which the reminders are stored
    """

    w_remind_keywords = {"w-remind": "en",
                         "w-erinner": "de"}
    """
    Keywords for the w-remind command
    """

    list_keywords = {"list": "en",
                     "liste": "de"}
    """
    Keywords for the list command
    """

    delete_keywords = {"delete": "en",
                       "löschen": "de"}
    """
    Keywords for the delete command
    """

    weekday_keywords = {"monday": ("en", "monday", 1),
                        "montag": ("de", "monday", 1),
                        "tuesday": ("en", "tuesday", 2),
                        "dienstag": ("de", "tuesday", 2),
                        "wednesday": ("en", "wednesday", 3),
                        "mittwoch": ("de", "wednesday", 3),
                        "thursday": ("en", "thursday", 4),
                        "donnerstag": ("de", "thursday", 4),
                        "friday": ("en", "friday", 5),
                        "freitag": ("de", "friday", 5),
                        "saturday": ("en", "saturday", 6),
                        "samstag": ("de", "saturday", 6),
                        "sunday": ("en", "sunday", 7),
                        "sonntag": ("de", "sunday", 7)}
    """
    Keywords to identify the individual weekdays
    """

    message_stored_message = {"en": "Reminder successfully stored",
                              "de": "Erinnerung erfolgreich abgespeichert"}
    """
    Message for the user letting him/her knoww that the reminder was stored
    """

    no_reminders_found_message = {"en": "No reminders stored",
                                  "de": "Keine Erinnerungen gespeichert"}
    """
    Message to be sent to the user when he/she requests the list of his/her reminders,
    but there are none saved
    """

    delete_index_out_of_bounds_message = {"en": "No reminder known for the selected index",
                                          "de": "Keine Erinnerung für den gegebenen Index bekannt"}
    """
    Message to be sent if the user's delete index is out of bounds
    """

    reminder_deleted_message = {"en": "Reminder successfully deleted",
                                "de": "Erinnerung erfolgreich gelöscht"}
    """
    Message to be sent if the user's delete request was successfull
    """

    def initialize(self) -> None:
        """
        Constructor extender for the WeeklyRenamer class that initializes a directory for the reminder files

        :return: None
        """
        self.w_reminder_directory = \
            os.path.join(LocalConfigChecker.services_directory, self.connection.identifier, self.w_reminder_directory)
        LocalConfigChecker.validate_directory(self.w_reminder_directory)

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        options, language, list_mode, delete_mode = self.parse_user_input(message.message_body)

        if list_mode:
            reply = self.list_stored_reminders(message.identifier, language)
        elif delete_mode:
            reply = self.delete_stored_reminder(message.identifier, language, options)
        else:
            reply = self.store_reminder(message.identifier, language, options[0], options[1], options[2])

        reply_message = self.generate_reply_message(message, "Weekly Reminder", reply)
        self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        weekdays = {}
        for weekday in WeeklyReminderService.weekday_keywords:
            weekdays[weekday] = WeeklyReminderService.weekday_keywords[weekday][0]

        base_regex = "^/" + Service.regex_string_from_dictionary_keys([WeeklyReminderService.w_remind_keywords]) + " "

        regex = base_regex + "\"[^\"]+\" " + Service.regex_string_from_dictionary_keys([weekdays])
        regex += " ([0-9]{2}-[0-9]{2}-[0-9]{2})$"

        list_regex = base_regex + Service.regex_string_from_dictionary_keys([WeeklyReminderService.list_keywords]) + "$"
        delete_regex = base_regex + Service.regex_string_from_dictionary_keys([WeeklyReminderService.delete_keywords])\
                                  + " [0-9]+$"

        return re.search(re.compile(regex), message.message_body.lower())\
            or re.search(re.compile(list_regex), message.message_body.lower())\
            or re.search(re.compile(delete_regex), message.message_body.lower())

    def parse_user_input(self, user_input: str) -> Tuple[(Tuple[str, str, str] or str), str, bool, bool]:
        """
        Parses the user input, figures out what the user wants, and the language used

        :param user_input: the user input to parse
        :return: the message to save as reminder together with the weekday and time (or the deletion index),
                    the language used, flag for list mode, flag for delete mode
        """
        list_mode = False
        delete_mode = False

        command_key = user_input.split(" ")[0].split("/")[1].lower()
        language = self.w_remind_keywords[command_key]

        options = user_input.lower().split(" ")
        options.pop(0)

        if len(options) == 1 and options[0] in self.list_keywords:
            list_mode = True
            return "", language, list_mode, delete_mode
        elif len(options) == 2 and options[0] in self.delete_keywords:
            delete_mode = True
            index = options[1]
            return index, language, list_mode, delete_mode
        else:
            message = user_input.split("\"", 1)[1].split("\"", 1)[0]
            options = user_input.lower().rsplit("\" ", 1)[1].split(" ")
            weekday = self.weekday_keywords[options[0]][1]
            time_of_day = options[1]
            return (message, weekday, time_of_day), language, list_mode, delete_mode

    def store_reminder(self, receiver: str, language: str, message: str, weekday: str, send_time: str) -> str:
        """
        Stores a continuous reminder as a local file

        :param receiver: the user requesting a new reminder
        :param language: The language for the reply message
        :param message: the message to be saved
        :param weekday: the weekday on which the reminder will be sent
        :param send_time: the time it will be sent
        :return: A message to be sent to the user
        """

        reminder_filename = ""
        index = 0

        # Emulate a do-while loop
        while True:
            reminder_filename = os.path.join(self.w_reminder_directory,
                                             weekday + "@" + send_time + "@" + receiver + "#####COUNT" + str(index))
            if not os.path.isfile(reminder_filename):
                break
            else:
                index += 1

        opened_file = open(reminder_filename, "w")
        opened_file.write(message)
        opened_file.close()

        return self.message_stored_message[language]

    def list_stored_reminders(self, sender: str, language: str) -> str:
        """
        Lists all stored reminders of a user

        :param sender: the user whoes reminders should be checked
        :param language: The language to be used when no entries were found
        :return: the list of reminders as a formatted string
        """
        list_of_reminders_string = ""
        list_of_reminders = self.get_user_reminders(sender)

        for index in range(0, len(list_of_reminders)):
            reminder = list_of_reminders[index]

            reminder_day = reminder["day"]
            for day in self.weekday_keywords:
                if self.weekday_keywords[day][1] == reminder_day and self.weekday_keywords[day][0] == language:
                    reminder_day = day

            list_of_reminders_string += str(index + 1) + ": " + reminder["send_time"] + " " + reminder_day + "\n"
            list_of_reminders_string += reminder["content"] + "\n\n"

        if list_of_reminders_string == "":
            list_of_reminders_string = self.no_reminders_found_message[language]

        return list_of_reminders_string.rsplit("\n\n", 1)[0]

    def delete_stored_reminder(self, sender: str, language: str, index: str) -> str:
        """
        Deletes a stored reminder

        :param sender:
        :param language:
        :param index:
        :return:
        """
        list_of_reminders = self.get_user_reminders(sender)
        index = int(index)

        if index < 1:
            return self.delete_index_out_of_bounds_message[language]

        try:
            reminder = list_of_reminders[index - 1]
            os.remove(reminder["file_path"])
            return self.reminder_deleted_message[language]

        except IndexError:
            return self.delete_index_out_of_bounds_message[language]

    def get_user_reminders(self, user: str) -> List[Dict[str, (str or int)]]:
        """
        Returns a list of reminders a user has currently stored

        :param user: the user whose reminders should be searched
        :return: A list of dictionaries with the information about the reminders
                    - day: number of day in week (e.g Monday=1) the message is usually sent
                    - send_time: at which time the reminder is usually sent
                    - content: The content of the reminder
                    - file_path: The path of the reminder file
        """
        list_of_reminders = []

        for file_name in os.listdir(self.w_reminder_directory):
            file_path = os.path.join(self.w_reminder_directory, file_name)

            if user in file_name:
                day = self.weekday_keywords[file_name.split("@", 1)[0]][1]
                send_time = file_name.split("@", 2)[1]

                opened_file = open(file_path, 'r')
                content = opened_file.read()
                opened_file.close()

                list_of_reminders.append({"day": day,
                                          "send_time": send_time,
                                          "content": content,
                                          "file_path": file_path})

        return sorted(list_of_reminders, key=lambda dictionary: (dictionary["day"], dictionary["send_time"]))

    def background_process(self) -> None:
        """
        Runs constantly in the background and checks if a message is due for sending

        :return: None
        """
        while True:

            # Calculate the current time
            current = datetime.datetime.utcnow()
            current_day = current.date().strftime("%A").lower()
            current_time = {"hour": current.hour,
                            "minute": current.minute,
                            "second": current.second}
            current_time_stamp = self.calculate_seconds_since_week_start(current_time, current_day)

            for file_name in os.listdir(self.w_reminder_directory):
                file_path = os.path.join(self.w_reminder_directory, file_name)

                # Calculate the due time for the reminder
                due_day = str(file_name.split("@", 1)[0])
                due_time = str(file_name.split("@", 2)[1])
                due_time = self.parse_time_string(due_time)
                reminder_timestamp = self.calculate_seconds_since_week_start(due_time, due_day)

                # Get the difference between the reminder time and the current time
                time_difference = abs(current_time_stamp - reminder_timestamp)

                # Send the reminder when in range of 3 seconds of the reminder time
                if time_difference < 3 and not file_name.endswith("+++WEEKDONE+++"):

                    # Read file content
                    opened_file = open(file_path, 'r')
                    content = opened_file.read()
                    opened_file.close()
                    os.rename(file_path, file_path + "+++WEEKDONE+++")  # Mark the file as done for the week

                    # Get the user's address
                    # noinspection PyTypeChecker
                    address = file_name.split("@", 2)[2].split("#####COUNT")[0]

                    message = Message(content, "Weekly Reminder", str(address), False)
                    self.send_text_message(message)

                if time_difference > 30 and file_name.endswith("+++WEEKDONE+++"):

                    os.rename(file_path, file_path.split("+++WEEKDONE+++")[0])  # Mark the file as not done

            time.sleep(1)

    @staticmethod
    def parse_time_string(time_string: str) -> Dict[str, int]:
        """
        Parses a time string to a dictionary

        :param time_string: the time string to parse
        :return: the time as a dictionary with the keys 'hour', 'minute' and 'second'
        """
        return {"hour": int(time_string.split("-")[0]),
                "minute": int(time_string.split("-")[1]),
                "second": int(time_string.split("-")[2])}

    def calculate_seconds_since_week_start(self, time_dictionary: Dict[str, int], weekday: str) -> int:
        """
        Calculates the amount of seconds that have passed for the specified time since the start of the week

        :param time_dictionary: the time as a dictionary
        :param weekday: the weekday to use in the calculation
        :return: the time in seconds since monday, 00:00:00
        """
        seconds = self.weekday_keywords[weekday][2] * 86400
        seconds += time_dictionary["second"]
        seconds += time_dictionary["minute"] * 60
        seconds += time_dictionary["hour"] * 3600
        return seconds



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import re

from kudubot.connection.generic.Message import Message
from kudubot.services.internet_services.CinemaService import CinemaService


class ZkmKinoService(CinemaService):
    """
    The ZkmKinoService Class that extends the generic Service class.
    The service fetches information for showtimes at the cinema:
    Filmpalast am ZKM in Karlsruhe
    """

    identifier = "zkm-kino"
    """
    The identifier for this service
    """

    help_description = {"en": "/zkm\tLists show time for the Filmpalast am ZKM\n"
                              "syntax:\n"
                              "/zkm <in how many days>",
                        "de": "/zkm\tListet Vorstelllungszeiten des Filmpalast am ZKM\n"
                              "syntax:\n"
                              "/zkm <in wievielen Tagen>"}
    """
    Help description for this service.
    """

    no_data_available = {"en": "No data available for this day (yet)",
                         "de": "(Noch) keine Daten für diesen Tag verfügbar"}

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        try:
            in_days = int(message.message_body.split("/zkm ")[1])
        except IndexError:
            in_days = 0
        cinema_data = self.get_cinema_data("karlsruhe", in_days, "f090a3fbcdb43f69")

        try:
            reply = self.format_cinema_data(cinema_data)[0]
        except IndexError:
            reply = self.no_data_available[self.connection.last_used_language]
            
        reply_message = self.generate_reply_message(message, "ZKM Kino", reply)
        self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        regex = "/zkm( [0-9]+)?"
        return re.search(re.compile(regex), message.message_body.lower())



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import re
import requests
from bs4 import BeautifulSoup
from typing import Tuple

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message


class KitMensaService(Service):
    """
    The KitMensaService Class that extends the generic Service class.
    The service fetches the current Mensa plan for the Mensa at the Karlsruhe
    Institute of Technology (KIT)
    """

    identifier = "kit_mensa"
    """
    The identifier for this service
    """

    help_description = {"en": "/mensa\tSends the Mensa Plan\n"
                              "syntax: /mensa [<line>] [tomorrow]",
                        "de": "/mensa\tSchickt den Mensa Plan\n"
                              "syntax: /mensa [<linie>] [morgen]"}
    """
    Help description for this service.
    """

    mensa_line = {"1": "",
                  "2": "",
                  "3": "",
                  "4": "",
                  "5": "",
                  "6": "",
                  "schnitzelbar": "",
                  "curry queen": "",
                  "abend": "",
                  "cafeteria vormittag": "",
                  "cafeteria nachmittag": ""}

    line_order = {0: "1",
                  1: "2",
                  2: "3",
                  3: "4",
                  4: "5",
                  5: "6",
                  6: "schnitzelbar",
                  7: "curry queen",
                  8: "abend",
                  9: "cafeteria vormittag",
                  10: "cafeteria nachmittag"}

    """
    Dictionary that keeps track of the offering at every line
    """

    line_key = {"line": "en",
                "linie": "de"}
    """
    The phrase for line in different languages
    """

    tomorrow_key = {"tomorrow": "en",
                    "morgen": "de"}
    """
    The phrase for tomorrow in different languages
    """

    closed = False
    """
    Set to True if the mensa is closed for the day
    """

    closed_message = {"en": "The mensa is closed today",
                      "de": "Die Mensa ist heute geschlossen"}
    """
    Message to be sent if the Mensa is closed
    """

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        language, line, tomorrow = self.parse_user_input(message.message_body.lower())
        self.get_current_info(tomorrow)

        if self.closed:
            reply = self.closed_message[language]

        elif line == "all":
            reply = ""
            first = False
            for line_no in range(0, len(self.mensa_line)):
                if first:
                    reply += self.mensa_line[self.line_order[line_no]]
                    first = False
                else:
                    reply += "\n" + self.mensa_line[self.line_order[line_no]]

        else:
            reply = self.mensa_line[line]

        reply_message = self.generate_reply_message(message, "KIT Mensa", reply)
        self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        regex = "^/(mensa)( )?(" + Service.regex_string_from_dictionary_keys([KitMensaService.line_key])
        regex += " (1|2|3|4|5|6)|schnitzelbar|curry queen|abend|cafeteria vormittag|cafeteria nachmittag)?( "
        regex += Service.regex_string_from_dictionary_keys([KitMensaService.tomorrow_key])
        regex += ")?$"

        return re.search(re.compile(regex), message.message_body.lower())

    def get_current_info(self, tomorrow: bool = False) -> None:
        """
        Fetches the current info either for today or tomorrow and stores them in a local dictionary

        :param tomorrow: can be set to True to get the information for tomorrow
        :return: None
        """
        try:

            if tomorrow:
                url = "http://mensa.akk.uni-karlsruhe.de/?DATUM=morgen&uni=1"
            else:
                url = "http://mensa.akk.uni-karlsruhe.de/?DATUM=heute&uni=1"

            html = requests.get(url).text
            soup = BeautifulSoup(html, "html.parser")
            resource = soup.select('body')
            body = resource[0].text

            self.mensa_line["1"] = "Linie 1\n" + body.split("Linie 1:\n", 1)[1].split("Linie 2:", 1)[0]
            self.mensa_line["2"] = "Linie 2\n" + body.split("Linie 2:\n", 1)[1].split("Linie 3:", 1)[0]
            self.mensa_line["3"] = "Linie 3\n" + body.split("Linie 3:\n", 1)[1].split("Linie 4/5:", 1)[0]
            self.mensa_line["4"] = "Linie 4/5\n" + body.split("Linie 4/5:\n", 1)[1].split("Schnitzelbar:", 1)[0]
            self.mensa_line["5"] = self.mensa_line["4"]
            self.mensa_line["6"] = "L6 Update\n" + body.split("L6 Update:\n", 1)[1].split("Abend:", 1)[0]
            self.mensa_line["abend"] = "Abend\n" + body.split("Abend:\n", 1)[1].split("Curry Queen:", 1)[0]
            self.mensa_line["schnitzelbar"] = \
                "Schnitzelbar\n" + body.split("Schnitzelbar:\n", 1)[1].split("L6 Update:", 1)[0]
            self.mensa_line["curry queen"] = \
                "Curry Queen\n" + body.split("Curry Queen:\n", 1)[1].split("Cafeteria Heiße Theke:", 1)[0]
            self.mensa_line["cafeteria vormittag"] = \
                "Cafeteria Heiße Theke\n" + \
                body.split("Cafeteria Heiße Theke:\n", 1)[1].split("Cafeteria ab 14:30:", 1)[0]
            self.mensa_line["cafeteria nachmittag"] = \
                "Cafeteria ab 14:30\n" + body.split("Cafeteria ab 14:30:\n", 1)[1].split("Stand:", 1)[0]

        except IndexError:
            self.closed = True

    def parse_user_input(self, user_input: str) -> Tuple[str, str, bool]:
        """
        Parses the user's input

        :param user_input: the input to be parsed
        :return: the language, the selected line, if today's or tomorrow's plan should be fetched
        """
        language = self.connection.last_used_language

        # Check if only /mensa is entered, if not, strip away /mensa_
        try:
            user_input = user_input.split("/mensa ")[1]
        except IndexError:
            return language, "all", False

        options = user_input.split(" ")  # Split up all words into seperate strings

        # Check if the plan for tomorrow should be fetched
        tomorrow = False
        for key in self.tomorrow_key:
            if key in options:
                tomorrow = True

        # If tomorrow is the only option, select all lines
        if options[0] in self.tomorrow_key:
            language = self.tomorrow_key[options[0]]
            return language, "all", True

        # Otherwise check which line
        else:
            if tomorrow:
                options.pop()  # Remove the tomorrow keyword
            line = ""
            for option in options:
                line += option + " "

            line = line.rstrip()

            for key in self.line_key:
                if key in line:
                    language = self.line_key[key]
                    line = line.replace(key + " ", "")

            return language, line, tomorrow



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import re

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message


class ServiceSelectorService(Service):
    """
    The ServiceSelectorService Class that extends the generic Service class.
    The service allows admin users to enable and disable services
    """

    identifier = "service_selector"
    """
    The identifier for this service
    """

    protected = True
    """
    May not be disabled
    """

    help_description = {"en": "/service\tallows activation/deactivation of services (admin)\n"
                              "/service activate <service|service-index>\tactivates a service\n"
                              "/service deactivate <service|service-index>\tdeactivates a service\n",
                        "de": "/service\tErmöglicht das Aktivieren/Deaktivieren von einem Service (admin)\n"
                              "/service an <service|service-index>\tAktiviert ein Service\n"
                              "/service aus <service|service-index>\tDeaktiviert ein Service\n"}
    """
    Help description for this service.
    """

    activation_keywords = {"activate": ("en", "activate"),
                           "deactivate": ("en", "deactivate"),
                           "an": ("de", "activate"),
                           "aus": ("de", "deactivate")}
    """
    Keywords that trigger a plugin activation or deactivation
    """

    already_activated = {"en": "is already activated or does not exist.",
                         "de": "ist bereits aktiviert oder existiert nicht."}
    """
    Reply for when a plugin was already activated
    """

    already_deactivated = {"en": "is already deactivated or does not exist.",
                           "de": "ist bereits deaktiviert oder existiert nicht."}
    """
    Reply for when a plugin was already deactivated
    """

    service_activated = {"en": "has been activated.",
                         "de": "wurde aktiviert."}
    """
    Reply for when a plugin was activated
    """

    service_deactivated = {"en": "has been deactivated.",
                           "de": "wurde deaktiviert."}
    """
    Reply for when a plugin was deactivated
    """

    protected_service_warning = {"en": "can not be deactivated",
                                 "de": "kann nicht deaktiviert werden"}
    """
    Reply when an attempt at deactivating a protected service is done
    """

    unauthorized_warning = {"en": "Sorry, I can't let you do that.",
                            "de": "Sorry, das darfst du nicht."}
    """
    Reply for non-admins
    """

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        authenticated = self.connection.authenticator.is_from_admin(message)

        parsed = message.message_body.lower().split("/service ", 1)[1]
        mode, selected_service_identifier = parsed.split(" ", 1)
        language = self.activation_keywords[mode][0]

        if authenticated:

            try:
                selected_service_identifier = \
                    self.connection.service_manager.all_services[int(selected_service_identifier) - 1].identifier
            except (IndexError, ValueError):
                pass

            selected_service_class = None
            for service in self.connection.service_manager.all_services:
                if service.identifier == selected_service_identifier:
                    selected_service_class = service

            is_active = False
            for service in self.connection.service_manager.active_services:
                if service.identifier == selected_service_identifier:
                    is_active = True

            if self.activation_keywords[mode][1] == "activate":
                if is_active or selected_service_class is None:
                    reply = selected_service_identifier + " " + self.already_activated[language]
                else:
                    self.connection.service_manager.active_services.append(selected_service_class)
                    reply = selected_service_identifier + " " + self.service_activated[language]
            else:
                if selected_service_class is None \
                        or selected_service_class not in self.connection.service_manager.active_services:
                    reply = selected_service_identifier + " " + self.already_deactivated[language]
                elif selected_service_class in self.connection.service_manager.protected_services:
                    reply = selected_service_identifier + " " + self.protected_service_warning[language]
                else:
                    self.connection.service_manager.active_services.remove(selected_service_class)
                    reply = selected_service_identifier + " " + self.service_deactivated[language]

        else:
            reply = self.unauthorized_warning[language]

        reply_message = self.generate_reply_message(message, "Help", reply)
        self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        regex = "^/service ("

        first = True
        for key in ServiceSelectorService.activation_keywords:
            if first:
                regex += key
                first = False
            else:
                regex += "|" + key

        regex += ") ([0-9]+|[a-zA-Z]+(( [a-zA-Z]+)+)?)$"

        return re.search(re.compile(regex), message.message_body.lower())



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import re
import requests
import urllib.request
import urllib.error
from typing import Tuple
from bs4 import BeautifulSoup

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message
from kudubot.config.LocalConfigChecker import LocalConfigChecker


class XkcdService(Service):
    """
    The XkcdService Class that extends the generic Service class.
    The service sends the comic strips from xkcd.com as images
    """

    identifier = "xkcd"
    """
    The identifier for this service
    """

    help_description = {"en": "/xkcd\tSends an XKCD comic strip\n"
                              "syntax:\n"
                              "/xkcd <comic_number>",
                        "de": "/xkcd\tSchickt einen XKCD Comic\n"
                              "Syntax:\n"
                              "/xkcd <comic_nummer>"}
    """
    Help description for this service.
    """

    xkcd_does_not_exist_message = {"en": "This comic does not exist (yet)",
                                   "de": "Dieser Comic existiert (noch) nicht"}

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        language = self.connection.last_used_language
        current_xkcd_comic = self.get_current_xkcd()

        try:
            selected_xkcd_comic = int(message.message_body.lower().split("/xkcd ")[1])
        except IndexError:
            selected_xkcd_comic = current_xkcd_comic

        if selected_xkcd_comic > current_xkcd_comic:
            reply = self.xkcd_does_not_exist_message[language]
            reply_message = self.generate_reply_message(message, "XKCD", reply)
            self.send_text_message(reply_message)
        else:
            file_path, title, alt_text = self.download_xkcd_comic(selected_xkcd_comic)
            self.send_image_message(message.address, file_path, title + "\n\n" + alt_text)
            self.delete_file_after(file_path, 5)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        regex = "^/xkcd( [0-9]+)?$"
        return re.search(re.compile(regex), message.message_body.lower())

    @staticmethod
    def get_current_xkcd() -> int:
        """
        Checks which is the curently newest XKCD comic strip

        :return: the number of the most current XKCD comic
        """
        current_xkcd_url = "http://xkcd.com"
        current_xkcd_html = requests.get(current_xkcd_url).text
        xkcd_soup = BeautifulSoup(current_xkcd_html, 'html.parser')

        xkcd_body = xkcd_soup.select('body')[0].text
        current = xkcd_body.split("Permanent link to this comic: http://xkcd.com/")[1].split("/\n")[0]

        return int(current)

    @staticmethod
    def download_xkcd_comic(number: int) -> Tuple[str, str, str]:
        """
        Downloads an XKCD Comic, finds the comic title and alt-text and returns this information
        :param number: the XKCD comic number to download
        :return: file path to the downloaded file, the title of the comic, the alt-text of the comic
        """
        xkcd_url = "http://xkcd.com/" + str(number)
        xkcd_html = requests.get(xkcd_url).text
        xkcd_soup = BeautifulSoup(xkcd_html, 'html.parser')

        comic_info = str(xkcd_soup.find("div", {"id": "comic"}))
        title = comic_info.split("img alt=\"")[1].split("\"")[0]
        alt_text = comic_info.split("title=\"")[1].split("\"/>")[0]
        image_link = 'http:' + comic_info.split("src=\"")[1].split("\"")[0]

        image_file = os.path.join(LocalConfigChecker.program_directory, "temp_xkcd.png")

        Service.wait_until_delete(image_file, 5)

        try:
            urllib.request.urlretrieve(image_link, image_file)
        except (urllib.error.HTTPError, urllib.error.URLError):
            pass

        return image_file, title, alt_text



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import re
import requests
from bs4 import BeautifulSoup
from urllib.parse import urlencode
from typing import Tuple, List, Dict

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message


class CinemaService(Service):
    """
    The CinemaService Class that extends the generic Service class.
    The service parses http://www.google.com/movies for current cinema show times
    and listings
    """

    identifier = "cinema"
    """
    The identifier for this service
    """

    help_description = {"en": "/cinema\tLists show times for movies in cinema\n"
                              "syntax:\n"
                              "/cinema <city> [in how many days]",
                        "de": "/kino\tListet die Spielzeiten für Filme im Kino\n"
                              "syntax:\n"
                              "/kino <stadt> [in wievielen Tage]"}
    """
    Help description for this service.
    """

    cinema_keywords = {"/cinema": "en",
                       "/kino": "de"}
    """
    Keywords for the cinema command
    """

    no_information_available = {"en": "Sorry, no information available",
                                "de": "Sorry, keine Daten verfügbar"}
    """
    Message sent when no information for the specified parameters could be found
    """

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        city, in_days = self.parse_user_input(message.message_body.lower())
        cinema_data = self.get_cinema_data(city, in_days)

        replies = self.format_cinema_data(cinema_data)

        if len(replies) == 0:
            reply = self.no_information_available[self.connection.last_used_language]
            reply_message = self.generate_reply_message(message, "Cinema", reply)
            self.send_text_message(reply_message)

        # We don't want to spam, so a current workaround is to only show largest theater
        largest_reply = ""
        for reply in replies:
            if len(reply) > len(largest_reply):
                largest_reply = reply
        reply_message = self.generate_reply_message(message, "Cinema", largest_reply)
        self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        regex = "^" + CinemaService.regex_string_from_dictionary_keys([CinemaService.cinema_keywords])
        regex += " (([a-zA-Z]{1}(([a-zA-Z]| )*))?[a-zA-Z]{1})( [0-9]+)?$"
        return re.search(re.compile(regex), message.message_body.lower())

    def parse_user_input(self, user_input: str) -> Tuple[str, int]:
        """
        Parses the user input for the city and the date for which information
        should be fetched

        :param user_input: the user input to parse
        :return: the city and how many days into the future the request is made
        """
        language_key, options = user_input.split(" ", 1)
        self.connection.last_used_language = self.cinema_keywords[language_key]

        if re.search(r"^(([^ ,]{1}([^,]*))?[^ ,]{1})( [0-9]+)+$", options):
            city, in_days = options.rsplit(" ", 1)
            return city, int(in_days)

        else:
            return options, 0

    # noinspection PyTypeChecker
    @staticmethod
    def format_cinema_data(cinema_data: Dict[str, Dict[str, (str or List[str])]]) -> List[str]:
        """
        Formats cinema data to be human-readable

        :param cinema_data: the cinema data
        :return: A list of formatted string, one for each theater
        """
        theaters = []

        for theater in cinema_data:
            theater_string = theater + "\n\n\n"
            for movie in cinema_data[theater]:
                theater_string += movie + "\n"
                theater_string += cinema_data[theater][movie]["runtime"] + "\n"
                for show_time in cinema_data[theater][movie]["times"]:
                    theater_string += show_time + " "
                theater_string = theater_string.rstrip()
                theater_string += "\n\n"
            theaters.append(theater_string.rstrip())

        return theaters

    # noinspection PyTypeChecker
    @staticmethod
    def get_cinema_data(city: str, in_days: int = 0, theater_override: str = "") \
            -> Dict[str, Dict[str, (str or List[str])]]:
        """
        Retrieves the data for the given parameters from google.com/movies

        :param city: the city whose cinemas should be listed
        :param in_days: the data from how many days into the future is requested
        :param theater_override: Can be used to specify a specific cinema
        :return: a dictionary containing all theaters and the movies playing in them, as well
                    as some basic information on those movies as well as the show times
        """

        # tid == theater ID, mid == movie ID
        payload = {"near": city, "date": str(in_days)}
        if theater_override:
            payload["tid"] = theater_override
        payload = urlencode(payload)

        google_request = requests.get("http://google.com/movies?" + payload).text
        google_soup = BeautifulSoup(google_request, 'html.parser')

        theaters = {}
        theater_selection = google_soup.find_all('div', attrs={'class': 'theater'})

        for theater in theater_selection:
            theater_name = theater.div.h2.text

            theaters[theater_name] = {}
            movies = theater.find_all('div', {'class': 'movie'})

            for movie in movies:
                movie_name = movie.div.text

                theaters[theater_name][movie_name] = {"runtime": "",
                                                      "age_info": "",
                                                      "times": []}
                if movie.span.text:
                    theaters[theater_name][movie_name]["runtime"] = movie.span.text.split(" - Rated ")[0]
                    try:
                        theaters[theater_name][movie_name]["age_info"] = movie.span.text.split(" - Rated ")[1]
                    except IndexError:
                        pass

                times = movie.find_all('div', {'class': 'times'})
                for show_time in times:
                    theaters[theater_name][movie_name]["times"].append(show_time.text)

        return theaters



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import re

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message


class XkcdRngService(Service):
    """
    The XkcdRngService Class that extends the generic Service class.
    The service returns the number 4, as illustrated in XKCD comic #221
    """

    identifier = "xkcd_rng"
    """
    The identifier for this service
    """

    help_description = {"en": "/xkcd-rng\tGenerates a guaranteed random number\n"
                              "syntax:\n"
                              "/xkcd-rng [source]\n"
                              "(the option 'source' sends the source code of the function)",
                        "de": "/xkcd-rng\tGeneriert eine gaantiert zufällige Nummer\n"
                              "/syntax:\n"
                              "/xkcd-rng [quelle]\n"
                              "(Die Option 'quelle' verschickt den Quellcode der Funktion)"}
    """
    Help description for this service.
    """

    source_keywords = {"source": "en",
                       "quelle": "de"}
    """
    Keywords for the source option
    """

    source_code = "int getRandomNumber()\n" \
                  "    return 4;    \\\\chosen by fair dice roll.\n" \
                  "                 \\\\guaranteed to be random.\n" \
                  "}"
    """
    The source code of the 'random' function
    """

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        reply = "4" if message.message_body.lower() == "/xkcd-rng" else self.source_code

        if reply != "4":
            self.connection.last_used_language = self.source_keywords[message.message_body.lower().split(" ")[1]]

        reply_message = self.generate_reply_message(message, "XKCD RNG", reply)
        if reply != "4" and self.connection.identifier in ["telegram", "whatsapp"]:
            self.send_text_as_image_message(reply_message)
        else:
            self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        regex = "^/xkcd-rng( " + Service.regex_string_from_dictionary_keys([XkcdRngService.source_keywords]) + ")?$"
        return re.search(re.compile(regex), message.message_body.lower())



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import re

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message


class MuterService(Service):
    """
    The MuterService Class that extends the generic Service class.
    It allows an authenticated admin user to mute and unmute a connection
    """

    identifier = "muter"
    """
    The identifier for this service
    """

    protected = True
    """
    May not be disabled
    """

    help_description = {"en": "/mute\tmutes the whatsbot (admin)\n"
                              "/unmute\tunmutes the whatsbot (admin)\n",
                        "de": "/stumm\tStellt den Bot auf lautlos (admin)\n"
                              "/laut\tHolt den Bot wieder aus dem Lautlosmodus aus (admin)\n"}
    """
    Help description for this service.
    """

    muter_keywords = {"mute": ("en", "mute"),
                      "unmute": ("en", "unmute"),
                      "stumm": ("de", "mute"),
                      "laut": ("de", "unmute")}
    """
    Keywords that trigger the muting/unmuting
    """

    unauthorized_warning = {"en": "Sorry, I can't let you do that.",
                            "de": "Sorry, das darfst du nicht."}
    """
    Reply for non-admins
    """

    not_muted_warning = {"en": "Bot is not muted",
                         "de": "Bot ist nicht stumm"}
    """
    Reply when the connection is not muted but it is attempted to unmute it
    """

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        authenticated = self.connection.authenticator.is_from_admin(message)

        mode = message.message_body.lower().split("/")[1]
        language = self.muter_keywords[mode][0]

        mute_state = False

        if authenticated:

            if mode == "mute" and not self.connection.muted:
                mute_state = True
                reply = "🤐"
            elif mode == "mute" and self.connection.muted:
                mute_state = True
                reply = ""
            elif mode == "unmute" and self.connection.muted:
                reply = "👍🏻"
            else:
                reply = self.not_muted_warning[language]
        else:
            reply = self.unauthorized_warning[language]

        if not mute_state:
            self.connection.muted = False

        if reply:
            reply_message = self.generate_reply_message(message, "Help", reply)
            self.send_text_message(reply_message)

        self.connection.muted = mute_state

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        regex = "^/("

        first = True
        for key in MuterService.muter_keywords:
            if first:
                regex += key
                first = False
            else:
                regex += "|" + key

        regex += ")$"

        return re.search(re.compile(regex), message.message_body.lower())



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import re
import json
import urllib.error
import urllib.request
from typing import List, Dict, Tuple
from urllib.parse import quote_plus, urlencode

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message


class KvvService(Service):
    """
    The KvvService Class that extends the generic Service class.
    The service checks the curent timetables from the KVV Straßenbahn Network
    and displays the most relevant times
    """

    identifier = "kvv"
    """
    The identifier for this service
    """

    help_description = {"en": "/kvv\tShows KVV timetable information\n"
                              "syntax:\n"
                              "/kvv <station>",
                        "de": "/kvv\tZeigt KVV Zeiten an\n"
                              "syntax:\n"
                              "/kvv <station>"}
    """
    Help description for this service.
    """

    kvv_api_key = "377d840e54b59adbe53608ba1aad70e8"
    """
    The KVV API key
    """

    kvv_url = "http://live.kvv.de/webapp/"
    """
    The KVV URL
    """

    corrections = {"hauptbahnhof": "karlsruhe hauptbahnhof",
                   "vorplatz": "karlsruhe hbf vorplatz",
                   "karlruhe hauptbahnhof vorplatz": "karlsruhe hbf vorplatz",
                   "hauptbahnhof vorplatz": "karlsruhe hbf vorplatz"}
    """
    Corrections for the station parameter given by the user to attempt to deliver more relevant
    search results
    """

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        station = message.message_body.split(" ", 1)[1]
        reply = self.get_kvv_info(station)
        reply_message = self.generate_reply_message(message, "KVV", reply)
        self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        regex = "^/kvv "
        return re.search(re.compile(regex), message.message_body.lower())

    def get_kvv_info(self, station: str) -> str:
        """
        Gets the KVV information for a selected station

        :param station: the station for which information should be looked up
        :return: the information for that station
        """
        station_ids = self.get_location_ids(station)

        for station_id in station_ids:
            try:
                times, official_station_name = self.get_times_from_stop_id(station_id['id'])
                return self.format_timetable(times, official_station_name)
            except urllib.error.HTTPError:
                pass

    # noinspection PyDefaultArgument
    def query_kvv_webapp(self, path: str, options: Dict[str, str]={}) -> Dict[str, str]:
        """
        Queries the KVV Webapp for a specified URL and returns a JSON parsed dictionary

        :param path: the URL path to be queried
        :param options: Options to be sent with the query
        :return: the JSON parsed dictionary
        """
        options["key"] = self.kvv_api_key
        query_url = self.kvv_url + path + "?" + urlencode(options)

        query_request = urllib.request.Request(query_url)
        opened_query_url = urllib.request.urlopen(query_request)
        result = opened_query_url.read().decode("utf-8")

        return json.loads(result)

    def get_location_ids(self, station_name) -> List[Dict[str, str]]:
        """
        Searches for the location id of a station

        :param station_name: the name of the station
        :return: the station ID
        """
        if station_name.lower() in self.corrections:
            station_name = self.corrections[station_name.lower()]

        station_name = station_name.replace(" ", "_")

        location_search_path = "stops/byname/" + quote_plus(station_name)
        json_parsed_dictionary = self.query_kvv_webapp(location_search_path)

        # noinspection PyTypeChecker
        return json_parsed_dictionary["stops"]

    def get_times_from_stop_id(self, station_id: str) -> Tuple[List[Dict[str, str]], str]:
        """
        Searches for departure times at a KVV station

        :param station_id: The station ID of the station to be queried
        :return: a list of departure dictionaries as well as the official station name
        """
        timetable_path = "departures/bystop/" + station_id
        json_parsed_dictionary = self.query_kvv_webapp(timetable_path)

        return json_parsed_dictionary["departures"], json_parsed_dictionary["stopName"]

    @staticmethod
    def format_timetable(station_times: List[Dict[str, str]], station_name: str) -> str:
        """
        Formats the timetable with the queried information

        :param station_times: the list of departure times of the station
        :param station_name: the station name
        :return: A formated string of the timetable information
        """
        direction_1_departures = []
        direction_2_departures = []
        for departure in station_times:
            if departure['direction'] == '1':
                direction_1_departures.append(departure)
            elif departure['direction'] == '2':
                direction_2_departures.append(departure)
        all_directions = [direction_1_departures, direction_2_departures]

        formatted_string = "Abfahrten an der Haltestelle " + station_name + ":\n\n"

        for direction in all_directions:
            for departure in direction:
                formatted_string += departure["time"] + " "
                formatted_string += departure["route"] + " "
                formatted_string += departure["destination"] + "\n"
            formatted_string += "\n"

        return formatted_string.rsplit("\n\n", 1)[0]



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import sys
import configparser

from kudubot.logger.PrintLogger import PrintLogger
from kudubot.config.LocalConfigChecker import LocalConfigChecker


class TelegramConfigParser(object):
    """
    Class that handles the telegram configuration
    """

    blank_config_file_template = "[credentials]\n" \
                                 "api_key = "

    @staticmethod
    def parse_telegram_config(connection_identifier: str) -> str:
        """
        Parses the Telegram config file and generates credentials from it

        :param connection_identifier: The identifier string of the Connection type
        :return: The API key for use with the TelegramConnection class
        """
        telegram_config_file = os.path.join(LocalConfigChecker.config_directory, connection_identifier)

        # First read the current file contents and perform sanity checks
        config_file = open(telegram_config_file, 'r')
        contents = config_file.read()
        config_file.close()

        # Is the file empty or doesn't have a credentials section? If yes, create basic template and delete current file
        if contents == "" or "[credentials]" not in contents:
            config_file = open(telegram_config_file, 'w')
            config_file.write(TelegramConfigParser.blank_config_file_template)
            PrintLogger.print("Generated Telegram Config Template, please enter your credentials in the file.")
            PrintLogger.print("The file is located at " + telegram_config_file)
            sys.exit(1)

        config = configparser.ConfigParser()
        config.read(telegram_config_file)
        parsed_config = dict(config.items("credentials"))

        try:
            # Get the values from the config file
            api_key = parsed_config["api_key"]

            # Check that the API key is filled out
            if not api_key:
                raise ValueError

            # If all went well, return the credentials
            return api_key

        except (KeyError, ValueError):
            PrintLogger.print("Invalid Telegram config file loaded. Please correct this.")
            sys.exit(1)



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import time
import irc.bot
import irc.strings
import irc.client

from typing import Tuple

from kudubot.logger.PrintLogger import PrintLogger
from kudubot.connection.generic.Message import Message
from kudubot.connection.generic.Connection import Connection
from kudubot.connection.irc.parsers.IrcConfigParser import IrcConfigParser


class IrcConnection(irc.bot.SingleServerIRCBot, Connection):
    """
    Class that implements the connection to an IRC network
    """

    identifier = "irc"
    """
    A string identifier with which other parts of the program can identify the type of connection
    """

    channel = ""
    """
    The channel this bot listens in on.
    """

    def __init__(self, credentials: Tuple[str, str, str, str]) -> None:
        """
        Constructor for the IRC Connection class that initializes the SingleServerIRC Bot

        :param credentials: The credentials used to log in.
        :return: None
        """
        super().__init__([(credentials[1], int(credentials[3]))], credentials[0], credentials[0])
        self.channel = credentials[2]

        self.initialize()

    # noinspection PyMethodMayBeStatic
    def on_nicknameinuse(self, connection: irc.client.Connection, event: irc.client.Event) -> None:
        """
        Method called when the nickname is already in use. It tries logging in again with the same
        name appended with an underscore

        :param connection: the connection to the IRC server
        :param event: The event that trigered the method call
        :return: None
        """
        str(event)
        connection.nick(connection.get_nickname() + "_")

    def on_welcome(self, connection: irc.client.Connection, event: irc.client.Event) -> None:
        """
        Method called when the bot connects to the IRC server. It then automatically joins the
        specified channel

        :param connection: the connection to the IRC server
        :param event: The event that triggered the method call
        :return: None
        """
        str(event)
        connection.join(self.channel)

    def on_pubmsg(self, connection: irc.client.Connection, event: irc.client.Event) -> None:
        """
        Method called whenever a message from a channel is received. Converts the event into
        a message and calls on_incoming_message with it.

        :param connection: the connection to the IRC server
        :param event: The event that triggered the method call
        :return: None
        """
        str(connection)
        received_text = event.arguments[0]
        sender_name = event.source.nick

        message = Message(message_body=received_text, message_title="", address=self.channel, incoming=True,
                          name=sender_name, group=True, single_address=sender_name, single_name=sender_name,
                          timestamp=-1.0)

        self.on_incoming_message(message)

    def on_privmsg(self, connection: irc.client.Connection, event: irc.client.Event) -> None:
        """
        Method called whenever a message from a private chat is received. Converts the event into
        a message and calls on_incoming_message with it.

        :param connection: the connection to the IRC server
        :param event: The event that triggered the method call
        :return: None
        """
        str(connection)
        received_text = event.arguments[0]
        sender_name = event.source.nick

        message = Message(message_body=received_text, message_title="", address=sender_name, incoming=True,
                          name=sender_name, group=False, timestamp=-1.0)

        self.on_incoming_message(message)

    def send_text_message(self, message: Message) -> None:
        """
        Sends a text message to the receiver. Some services allow the use of titles, but some don't,
        so the message title is optional

        :param message: The message to be sent
        :return: None
        """
        command = self.connection.notice if message.group else self.connection.privmsg

        for line in message.message_body.split("\n"):
            command(message.address, line)
            time.sleep(0.3)

    def send_image_message(self, receiver: str, message_image: str, caption: str = "") -> None:
        """
        Sends an image to the receiver, with an optional caption/title

        :param receiver: The receiver of the message
        :param message_image: The image to be sent
        :param caption: The caption/title to be displayed along with the image, defaults to an empty string
        :return: None
        """
        pass

    def send_audio_message(self, receiver: str, message_audio: str, caption: str = "") -> None:
        """
        Sends an audio file to the receiver, with an optional caption/title

        :param receiver: The receiver of the message
        :param message_audio: The audio file to be sent
        :param caption: The caption/title to be displayed along with the audio, defaults to an empty string
        :return: None
        """
        pass

    @staticmethod
    def establish_connection() -> None:
        """
        Establishes the connection to the specific service

        :return: None
        """
        credentials = IrcConfigParser.parse_irc_config(IrcConnection.identifier)

        while True:
            bot = IrcConnection(credentials)
            PrintLogger.print("Starting IRC connection", 1)
            bot.start()



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import re
import pywapi
from typing import Tuple, Dict

from kudubot.servicehandlers.Service import Service
from kudubot.connection.generic.Message import Message


class WeatherService(Service):
    """
    The KickTippService Class that extends the generic Service class.
    The service parses www.kicktipp.de to get a kicktipp group's current standings
    """

    identifier = "weather"
    """
    The identifier for this service
    """

    help_description = {"en": "/weather\tSends weather information\n"
                              "syntax:\t/weather[:][options;] <cityname>[, <region>][, <country>]\n"
                              "options: text, verbose",
                        "de": "/wetter\tSchickt Wetterinformationen\n"
                              "syntax:\t/wetter[optionen;] <stadtname>[, <region>][, <land>]\n"
                              "optionen: text, verbos"}
    """
   Help description for this service.
   """

    weather_keywords = {"weather": "en",
                        "wetter": "de"}
    """
    Keywords that map to the language to be used
    """

    options = {"text": False,
               "verbose": False}

    weather_identifiers = {"sunny": {"em":  "☀", "de": "sonnig"},
                           "clear": {"em":  "☀", "de": "klar"},
                           "sunny / windy": {"em":  "☀", "de": "sonnig / windig"},
                           "clear / windy": {"em":  "☀", "de": "klar / windig"},
                           "light rain": {"em":  "🌧", "de": "leichter Regen"},
                           "light rain shower": {"em":  "🌧", "de": "leichter Regenschauer"},
                           "haze": {"em":  "🌫", "de": "Dunst"},
                           "fog": {"em":  "🌫", "de": "Nebel"},
                           "mist": {"em":  "🌫", "de": "Nebel"},
                           "thunderstorms": {"em":  "⛈", "de": "Gewitter"},
                           "t-storm": {"em":  "⛈", "de": "Gewitter"},
                           "fair": {"em":  "🌤", "de": "recht sonnig"},
                           "partly cloudy": {"em":  "⛅", "de": "teils bewölkt"},
                           "mostly cloudy": {"em":  "🌥", "de": "stark bewölkt"},
                           "not defined": {"em":  "❔", "de": "nicht definiert"},
                           "cloudy": {"em":  "☁", "de": "bewölkt"},
                           "rain shower": {"em":  "☔", "de": "Regenschauer"},
                           "thunderclouds": {"em":  "🌩", "de": "Gewitterwolken"},
                           "snow": {"em":  "🌨", "de": "Schnee"},
                           "windy": {"em":  "🌬", "de": "windig"},
                           "tornado": {"em":  "🌪", "de": "Tornado"}}
    """
    A dictionary of weather identifiers for icons (and different languages)
    """

    weather_message_dictionary = {"en": ("It is ", " and ", " now in "),
                                  "de": ("Es ist ", " und ", " in ")}
    """
    Defines the weather message for multiple languages
    """

    city_not_found_message = {"en": "City not found",
                              "de": "Stadt nicht gefunden"}

    def process_message(self, message: Message) -> None:
        """
        Process a message according to the service's functionality

        :param message: the message to process
        :return: None
        """
        language, city, country, region = self.parse_user_input(message)
        self.connection.last_used_language = language

        location = self.get_location(city, country, region)
        if location is not None:
            weather = pywapi.get_weather_from_weather_com(location[0])
            reply = self.generate_weather_string(language, location, weather)
        else:
            reply = self.city_not_found_message[language]

        reply_message = self.generate_reply_message(message, "Weather", reply)
        self.send_text_message(reply_message)

    @staticmethod
    def regex_check(message: Message) -> bool:
        """
        Checks if the user input is valid for this service to continue

        :return: True if input is valid, False otherwise
        """
        regex = "^/" + Service.regex_string_from_dictionary_keys([WeatherService.weather_keywords])
        regex += "(:(text;|verbose;)+)?( (([^ ,]+| )+)(, ([^ ,]+| )+)?(, ([^ ,]+| )+)?)?$"

        return re.search(re.compile(regex), message.message_body)

    def parse_user_input(self, message: Message) -> Tuple[str, str, str, str]:
        """
        Parses the user input. It determines the used language, if text mode or verbose mode should
        be used and the city, country and region to check
        The options are stored in the options dictionary

        :return: a tuple of the parsed information in the order:
                    -language
                    -city
                    -country
                    -region
        """
        message = message.message_body.lower().split("/", 1)[1].split(" ")
        command = message[0]
        option_split = command.split(":")

        language = self.weather_keywords[option_split[0]]

        verbose = False
        text_mode = False

        if len(option_split) > 1:
            options = option_split[1].split(";")
            if "verbose" in options:
                verbose = True
            if "text" in options:
                text_mode = True

        self.options["text"] = text_mode
        self.options["verbose"] = verbose

        city = ""
        country = ""
        region = ""

        # This concats all space-delimited parts of the string for so long until a comma is found at the end,
        # then it jumps on to the next-less relevant regional identifier (city->country->region)
        for position in range(1, len(message)):
            if city.endswith(","):
                if country.endswith(","):
                    if region.endswith(","):
                        break
                    else:
                        region += " " + message[position]
                else:
                    country += " " + message[position]
            else:
                city += " " + message[position]

        # Remove commas and surrounding space characters
        city = city.replace(",", "").strip()
        country = country.replace(",", "").strip()
        region = region.replace(",", "").strip()

        return language, city, country, region

    @staticmethod
    def get_location(city: str, country: str, region: str) -> Tuple[str, str]:
        """
        Gets a pywapi location ID Tuple for a city, country and region

        :param city: The city to be searched
        :param country: The country to be searched
        :param region: The region to be searched
        :return: the pywapi location ID Tuple, or None if no location was found
        """
        search_term = city
        if country:
            search_term += (", " + country)
        if region:
            search_term += (", " + region)

        try:
            location = pywapi.get_loc_id_from_weather_com(search_term)[0]
        except KeyError:
            return None

        # Add 'USA' to US-american location strings
        if len(location[1].split(", ")) < 3:
            location_string = location[1]
            location_string += ", USA"
            location = (location[0], location_string)

        return location

    def generate_weather_string(self, language: str, location: Tuple[str, str], weather: Dict) -> str:
        """
        Generates a message string to send back for the selected location's weather using the
        specified options

        :param language: the language to use
        :param location: the loaction tuple of the location to check
        :param weather: the weather dictionary of that location
        :return: the message string
        """
        try:
            weather_type = weather['current_conditions']['text']
            temperature = weather['current_conditions']['temperature']
            location_string = location[1]

            if not self.options['verbose']:
                location_string = location_string.split(", ")[0] + ", " + location_string.split(", ")[2]

            if not self.options['text']:
                try:
                    weather_type = WeatherService.weather_identifiers[weather_type.lower()]["em"]
                except KeyError:
                    weather_type = WeatherService.weather_identifiers["not defined"]["em"]
            else:
                if language != "en":
                    try:
                        weather_type = WeatherService.weather_identifiers[weather_type.lower()]["de"]
                    except KeyError:
                        weather_type = WeatherService.weather_identifiers["not defined"]["de"]

            language_message = WeatherService.weather_message_dictionary[language]
            message_string = language_message[0] + weather_type + language_message[1]
            message_string += temperature + "°C" + language_message[2] + location_string

            return message_string
        except KeyError:
            return "Weather data currently unavailable"



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import sys

from yowsup.layers.interface import YowInterfaceLayer
from yowsup.layers.interface import ProtocolEntityCallback
from yowsup.layers.protocol_media.mediauploader import MediaUploader
from yowsup.layers.protocol_messages.protocolentities import MessageProtocolEntity
from yowsup.layers.protocol_messages.protocolentities import TextMessageProtocolEntity
from yowsup.layers.protocol_media.protocolentities import RequestUploadIqProtocolEntity
from yowsup.layers.protocol_media.protocolentities import ResultRequestUploadIqProtocolEntity
from yowsup.layers.protocol_media.protocolentities import ImageDownloadableMediaMessageProtocolEntity
from yowsup.layers.protocol_media.protocolentities import AudioDownloadableMediaMessageProtocolEntity
from yowsup.layers.protocol_media.protocolentities import VideoDownloadableMediaMessageProtocolEntity
from yowsup.layers.protocol_media.protocolentities.builder_message_media_downloadable \
    import DownloadableMediaMessageBuilder


from kudubot.logger.PrintLogger import PrintLogger


class YowsupEchoLayer(YowInterfaceLayer):
    """
    The Yowsup Echo layer class
    The layer component of the whatsapp connection. Used to receive messages.
    """

    # Required local variables
    disconnect_action_prompt = 0
    accountDelWarnings = 0
    connected = False
    username = None
    sendReceipts = True
    disconnectAction = 0
    credentials = None
    jid_aliases = {}

    @ProtocolEntityCallback("receipt")
    def on_receipt(self, entity: MessageProtocolEntity) -> None:
        """
        method run whenever a whatsapp receipt is issued

        :param entity: The receipt entity
        :return: void
        """
        self.toLower(entity.ack())

    def send_media(self, number: str, path: str, media_type: RequestUploadIqProtocolEntity.TYPES_MEDIA,
                   caption: str = None) -> None:
        """
            Sends a media file

            :param number: the receiver of the media file
            :param path: the path to the media file
            :param media_type: The type of media to send
            :param caption: the caption to be shown
            :return: void
            """
        jid = self.alias_to_jid(number)
        entity = RequestUploadIqProtocolEntity(media_type, filePath=path)

        def success_fn(success_entity: TextMessageProtocolEntity, original_entity: MessageProtocolEntity) -> None:
            """
            Function called on successful media upload

            :param success_entity: the success entity
            :param original_entity: the original entity
            :return: None
            """
            # noinspection PyTypeChecker
            self.on_request_upload_result(jid, media_type, path, success_entity, original_entity, caption)

        def error_fn(error_entity: MessageProtocolEntity, original_entity: RequestUploadIqProtocolEntity) -> None:
            """
            Function called on failed media upload

            :param error_entity: the error entity
            :param original_entity: the original entity
            :return: None
            """
            YowsupEchoLayer.on_request_upload_error(jid, path, error_entity, original_entity)

        self._sendIq(entity, success_fn, error_fn)

    def send_image(self, number: str, path: str, caption: str = None) -> None:
        """
        Sends an image

        :param number: the receiver of the image
        :param path: the path to the image file
        :param caption: the caption to be shown
        :return: void
        """
        self.send_media(number, path, RequestUploadIqProtocolEntity.MEDIA_TYPE_IMAGE, caption)

    def send_audio(self, number: str, path: str) -> None:
        """
        Sends an audio file

        :param number: the number of the receiver of the file
        :param path: the path to the audio file
        :return: None
        """
        self.send_media(number, path, RequestUploadIqProtocolEntity.MEDIA_TYPE_AUDIO)

    def send_video(self, number: str, path: str, caption: str = None) -> None:
        """
        Sends a video file

        :param number: the number of the receiver of the file
        :param path: the path to the video file
        :param caption: an optional caption to be displayed together with the video file
        :return: None
        """
        self.send_media(number, path, RequestUploadIqProtocolEntity.MEDIA_TYPE_VIDEO, caption)

    def do_send_media(self, media_type: RequestUploadIqProtocolEntity.TYPES_MEDIA,
                      file_path: str, url: str, to: str, ip: str = None, caption: str = None) -> None:
        """
        Sends a media file

        :param media_type: The type of media to send
        :param file_path: the path to the file
        :param url: the whatsapp upload url
        :param to: the receiver
        :param ip: the ip of the receiver
        :param caption: the caption to be displayed together with the media file
        :return: None
        """
        entity = None
        if media_type == RequestUploadIqProtocolEntity.MEDIA_TYPE_IMAGE:
            entity = ImageDownloadableMediaMessageProtocolEntity.fromFilePath(file_path, url, ip, to, caption=caption)
        elif media_type == RequestUploadIqProtocolEntity.MEDIA_TYPE_AUDIO:
            entity = self.create_audio_media(file_path, url, to, ip)
        elif media_type == RequestUploadIqProtocolEntity.MEDIA_TYPE_VIDEO:
            entity = VideoDownloadableMediaMessageProtocolEntity.fromFilePath(file_path, url, ip, to, caption=caption)
        self.toLower(entity)

    @staticmethod
    def create_audio_media(file_path: str, url: str, to: str, ip: str = None) \
            -> AudioDownloadableMediaMessageProtocolEntity:
        """
        Creates an Audio Message Entity as a workaround for the missing fromFilePath method in
        AudioDownloadableMediaMessageProtocolEntity's parent class

        :param file_path: the path to the file
        :param url: the whatsapp upload url
        :param to: the receiver
        :param ip: the ip of the receiver
        :return: None
        :return: the generatd audio entity
        """
        builder = DownloadableMediaMessageBuilder(AudioDownloadableMediaMessageProtocolEntity, to, file_path)
        builder.set("url", url)
        builder.set("ip", ip)
        return AudioDownloadableMediaMessageProtocolEntity.fromBuilder(builder)

    def on_request_upload_result(self, jid: str, media_type: RequestUploadIqProtocolEntity.TYPES_MEDIA, file_path: str,
                                 result_request_upload_iq_protocol_entity: ResultRequestUploadIqProtocolEntity,
                                 request_upload_iq_protocol_entity: RequestUploadIqProtocolEntity,
                                 caption: str = None) -> None:
        """
        Method run when a media upload result is positive

        :param jid: the jid to receive the media
        :param media_type: The media type to send
        :param file_path: the path to the media file
        :param result_request_upload_iq_protocol_entity: the result entity
        :param request_upload_iq_protocol_entity: the request entity
        :param caption: the media caption, if applicable
        :return: None
        """
        str(request_upload_iq_protocol_entity)
        if result_request_upload_iq_protocol_entity.isDuplicate():
            self.do_send_media(media_type, file_path, result_request_upload_iq_protocol_entity.getUrl(), jid,
                               result_request_upload_iq_protocol_entity.getIp(), caption)
        else:

            def success_fn(inner_file_path: str, inner_jid: str, url: str) -> None:
                """
                Function called when upload was successful

                :param inner_file_path: path to the media file
                :param inner_jid: the receiver's jid
                :param url: the whatsapp media url
                :return: None
                """
                self.do_send_media(media_type, inner_file_path, url, inner_jid,
                                   result_request_upload_iq_protocol_entity.getIp(), caption)

            media_uploader = MediaUploader(jid, self.getOwnJid(), file_path,
                                           result_request_upload_iq_protocol_entity.getUrl(),
                                           result_request_upload_iq_protocol_entity.getResumeOffset(), success_fn,
                                           YowsupEchoLayer.on_upload_error, YowsupEchoLayer.on_upload_progress,
                                           async=False)
            media_uploader.start()

    @staticmethod
    def on_request_upload_error(jid: str, path: str, error_request_upload_iq_protocol_entity: MessageProtocolEntity,
                                request_upload_iq_protocol_entity: RequestUploadIqProtocolEntity) -> None:
        """
        Method run when a media upload result is negative

        :param jid: the jid to receive the media
        :param path: the file path to the media
        :param error_request_upload_iq_protocol_entity: the error result entity
        :param request_upload_iq_protocol_entity: the request entity
        :return: None
        """
        if error_request_upload_iq_protocol_entity and request_upload_iq_protocol_entity:
            PrintLogger.print("Request upload for file %s for %s failed" % (path, jid))

    @staticmethod
    def on_upload_error(file_path: str, jid: str, url: str) -> None:
        """
        Method run when an upload error occurs

        :param file_path: the file path of the file to upload
        :param jid: the jid of the receiver of the file
        :param url: the upload url
        :return: None
        """
        PrintLogger.print("Upload file %s to %s for %s failed!" % (file_path, url, jid))

    @staticmethod
    def on_upload_progress(file_path: str, jid: str, url: str, progress: float) -> None:
        """
        Method that keeps track of the upload process

        :param file_path: the file path of the media file
        :param jid: the jid of the receiver
        :param url: the whatsapp upload url
        :param progress: the current progress
        :return:None
        """
        if url:
            sys.stdout.write("%s => %s, %d%% \r" % (os.path.basename(file_path), jid, progress))
            sys.stdout.flush()

    def alias_to_jid(self, c_alias: str) -> str:
        """
        Turns an alias into a jid

        :param c_alias: the current alias
        :return: the jid
        """
        for alias, a_jid in self.jid_aliases.items():
            if c_alias.lower() == alias.lower():
                return YowsupEchoLayer.normalize_jid(a_jid)

        return YowsupEchoLayer.normalize_jid(c_alias)

    @staticmethod
    def normalize_jid(number: str) -> str:
        """
        Normalizes a jid

        :param number: the number to be receive a normalized jid
        :return: the normalized jid
        """
        if '@' in number:
            return number
        elif "-" in number:
            return number + "@g.us"

        return number + "@s.whatsapp.net"

    def send_receipt(self, message_protocol_entity: MessageProtocolEntity) -> None:
        """
        Sends the whatsapp servers that the message was received

        :param message_protocol_entity: the message protocol entity that was received
        :return: None
        """
        self.toLower(message_protocol_entity.ack())
        self.toLower(message_protocol_entity.ack(True))



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
from typing import Tuple

from kudubot.connection.generic.Message import Message
from kudubot.connection.generic.Connection import Connection
from kudubot.connection.email.senders.SmtpSender import SmtpSender
from kudubot.connection.email.listeners.ImapListener import ImapListener
from kudubot.connection.email.parsers.EmailConfigParser import EmailConfigParser


class EmailConnection(Connection):
    """
    Class that implements an Email-based connection using imaplib and smtplib
    """

    identifier = "email"
    """
    A string identifier with which other parts of the program can identify the type of connection
    """

    credentials = ()
    """
    The credentials used to connect to the IMAP and SMTP servers
    The are a tuple of the form (email address, password, server, imap port, smtp port)
    """

    smtp_sender = None
    """
    An SMTP handler that can be used to send email messages
    """

    def __init__(self, credentials: Tuple[str, str, str, str, str]) -> None:
        """
        Constructor for the EmailConnection class. It stores the credentials and generates
        an SMTP connection handler

        :param credentials: The credentials used to connect to the email server:
                                (email address, password, server, imap port, smtp port)
        :return: None
        """
        self.initialize()
        self.credentials = credentials
        self.smtp_sender = SmtpSender(credentials)

    def send_text_message(self, message: Message) -> None:
        """
        Sends a text message to the receiver.

        :param message: The message entity to be sent
        :return: None
        """
        self.smtp_sender.send_text_email(message)

    def send_image_message(self, receiver: str, message_image: str, caption: str = "") -> None:
        """
        Sends an image to the receiver, with an optional caption/title

        :param receiver: The receiver of the message
        :param message_image: The image to be sent
        :param caption: The caption/title to be displayed along with the image, defaults to an empty string
        :return: None
        """
        self.smtp_sender.send_image_email(message_image, caption, receiver)

    def send_audio_message(self, receiver: str, message_audio: str, caption: str = "") -> None:
        """
        Sends an audio file to the receiver, with an optional caption/title

        :param receiver: The receiver of the message
        :param message_audio: The audio file to be sent
        :param caption: The caption/title to be displayed along with the audio, defaults to an empty string
        :return: None
        """
        self.smtp_sender.send_audio_email(message_audio, caption, receiver)

    @staticmethod
    def establish_connection() -> None:
        """
        Establishes the connection to the specific service

        :return: None
        """
        credentials = EmailConfigParser.parse_email_config(EmailConnection.identifier)
        email_connection = EmailConnection(credentials)
        ImapListener(credentials, email_connection.on_incoming_message).listen()



# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

verbosity = 0
"""
Identifier for the selected verbosity
"""

"""
The metadata is stored here. It can be used by any other module in this project this way, most
notably by the setup.py file
"""

project_name = "kudubot"
"""
The name of the project
"""

project_description = "A bot that interfaces with several different messenger services"
"""
A short description of the project
"""

version_number = "0.7.5"
"""
The current version of the program.
"""

development_status = "Development Status :: 3 - Alpha"
"""
The current development status of the program
"""

project_url = "http://namibsun.net/namboy94/kudubot"
"""
A URL linking to the home page of the project, in this case a
self-hosted Gitlab page
"""

download_url = "http://gitlab.namibsun.net/namboy94/kudubot/repository/archive.zip?ref=master"
"""
A URL linking to the current source zip file.
"""

author_name = "Hermann Krumrey"
"""
The name(s) of the project author(s)
"""

author_email = "hermann@krumreyh.com"
"""
The email address(es) of the project author(s)
"""

license_type = "GNU GPL3"
"""
The project's license type
"""

dependencies = ['tvdb_api',
                'yowsup2',
                'pywapi',
                'pillow',
                'beautifulsoup4',
                'typing',
                'python-telegram-bot',
                'gTTS',
                'irc']
"""
Python Packaging Index requirements
"""

audiences = ["Intended Audience :: End Users/Desktop",
             "Intended Audience :: Developers"]
"""
The intended audience of this software
"""

environment = "Environment :: Other Environment"
"""
The intended environment in which the program will be used
"""

programming_languages = ['Programming Language :: Python :: 3',
                         'Programming Language :: Python :: 2']
"""
The programming language used in this project
"""

topic = "Topic :: Internet"
"""
The broad subject/topic of the project
"""

language = "Natural Language :: English"
"""
The (default) language of this project
"""

compatible_os = "Operating System :: POSIX :: Linux"
"""
The Operating Systems on which the program can run
"""

license_identifier = "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
"""
The license used for this project
"""


# coding=utf-8
"""
LICENSE:
Copyright 2015,2016 Hermann Krumrey

This file is part of kudubot.

    kudubot makes use of various third-party python modules to serve
    information via online chat services.

    kudubot is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    kudubot is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import sys
import argparse
import traceback
from threading import Thread

import kudubot.metadata as metadata
from kudubot.logger.PrintLogger import PrintLogger
from kudubot.logger.ExceptionLogger import ExceptionLogger
from kudubot.config.LocalConfigChecker import LocalConfigChecker
from kudubot.connection.email.EmailConnection import EmailConnection
from kudubot.connection.whatsapp.WhatsappConnection import WhatsappConnection
from kudubot.connection.telegram.TelegramConnection import TelegramConnection

connections = [EmailConnection,
               WhatsappConnection,
               TelegramConnection]
"""
A list of possible connections
"""


def main(override: str = "", verbosity: int = 1) -> None:
    """
    The main method of the program

    :param override: Can be used to override the main method to force a specific connection to run
    :param verbosity: Can be set to define how verbose the outpt will be. Defaults to 0, no or only basic output
    :return: None
    """
    PrintLogger.print("Messengerbot V " + metadata.version_number)
    PrintLogger.print("Parsing Command Line Arguments", 5)

    if override and len(sys.argv) == 1:
        sys.argv.append(override)

    parser = argparse.ArgumentParser()
    parser.add_argument("mode", help="The connection type to start. Can be 'email', 'telegram', or 'all'")
    parser.add_argument("--verbosity", help="Sets the output verbosity", type=int)
    args = parser.parse_args()

    metadata.verbosity = args.verbosity if args.verbosity else verbosity

    # block stderr messages from being printed.
    dev_null = open(os.devnull, 'w')
    sys.stderr = dev_null

    PrintLogger.print("Starting program", 1)

    try:
        try:
            # Check if the local configs are OK and if necessary fix them
            LocalConfigChecker.check_and_fix_config(connections)

            if args.mode == "all":
                for connection in connections:

                    def connect():
                        """
                        Connects to a service

                        :return: None
                        """
                        try:
                            connection.establish_connection()
                        except Exception as ex:
                            stack_t = traceback.format_exc()
                            ExceptionLogger.log_exception(ex, stack_t, connection.identifier)

                    connection_thread = Thread(target=connect)
                    connection_thread.daemon = True
                    connection_thread.start()
                while True:
                    pass
            else:
                # Generate the connection
                connected = False
                for connection in connections:
                    if connection.identifier == args.mode:
                        connected = True
                        connection.establish_connection()

                if not connected:
                    PrintLogger.print("No valid connection type selected")
                    sys.exit(1)
        except Exception as e:
            stack_trace = traceback.format_exc()
            ExceptionLogger.log_exception(e, stack_trace, "main")

    except KeyboardInterrupt:
        pass

    PrintLogger.print("Thanks for using kudubot")


if __name__ == "__main__":
    main()
